% File:   LLR_Advantage_16QAM_LTE.m 
% Author: Andreas Schwarzinger                                                     Date: 10.5.2020
% Notes:  The aim of this script is to illustrate the advantage of LLR scaling ahead of binary convolutional
%         decoding.

clc;
clear variables;
close all;

%% 0. Some definitions regarding the convolutional encoder
%     The specifications used here are those of the LTE PBCH BCC encoder (See 3GPP TS36.212 Sect. 5.1.3.1

QamIndex             = 6; % BPSK = 1, QPSK = 2, 16QAM = 4, 64QAM = 6
rng(0);                   % Set the random generator seed to 0 to get repeatable bits.

DecisionMethod       = 'Hard';   % 'Soft', 'LLR', 'Hard'
DecisionMethods      = cell(3,1);
DecisionMethods{1,1} = 'Hard';
DecisionMethods{2,1} = 'Soft';
DecisionMethods{3,1} = 'LLR';

ConstraintLength = 7;  
trellis          = poly2trellis(ConstraintLength,[133,171,165]);
InitState        = 0;

Noise_Power  = 0.25;
NumRuns      = 25;
SNRdB_Array  = zeros(3, NumRuns); 
BER_Array    = zeros(3, NumRuns);

for DecisionIndex = 1:length(DecisionMethods)
    DecisionMethod = DecisionMethods{DecisionIndex, 1};
    Noise_Power  = 0.1;
    for Run = 1:NumRuns
        Noise_Power = Noise_Power*1.2;
%    SNR_dB           = 0;
%    SNR_Linear       = 10^(SNR_dB/10);  
%    Signal_Power     = 1;               % As the BPSK symbols are equal to +/-1   +  j0
%    Noise_Power      = Signal_Power/SNR_Linear;
    

        %% 1. Define the input bit sequence
        NumberOfInfoBit     = 30000 * QamIndex;   % Ensure that we have the right number of bits in the 
                                                  % message for the QamIndex we are using
        InfoBits            = randi([0, 1], 1, NumberOfInfoBit);
        NumberOfPadBits     = 12;                 % This number of pad bits will result in a valid number
                                                  % of QAM symbols.
                                                  % was --> ConstraintLength - 1;
        PadBits             = zeros(1, NumberOfPadBits);
        BitMessage          = [InfoBits, PadBits];
        assert(sum(mod(length(BitMessage), [1, 2, 4, 6]) == [0, 0, 0, 0]) == 4, 'There is a problem with the message length.');
        

        %% 2. Encode the BitMessage
        %     The encoding step            
        [EncodedBits]    = convenc(BitMessage, trellis, InitState);

        %% 2a  Convert to inverted bipolar --> Map bit value = 0 to +1.0,   and a bit value = 1 to -1.0
        BPSK_Message        = -(EncodedBits*2 - 1); 
        ResizedBitStream    = reshape(EncodedBits, QamIndex, []);
        NumberOfInfoSymbols = size(ResizedBitStream, 2);
        QamMessage          = zeros(1, NumberOfInfoSymbols);
        for i = 1:NumberOfInfoSymbols
             QamMessage(1, i) = LteQamMapper(ResizedBitStream(:, i));
        end
        
        

        %% 3. Distort via Channel Response and Noise
        ChannelResponse           = randi([30, 500], 1, length(QamMessage))/100;
        for i = 1:length(ChannelResponse)
            if(ChannelResponse(1,i) > 1); ChannelResponse(1,i) = 1; end
        end
        Qam_Message_AfterChannel = QamMessage.*ChannelResponse;
    

        %% 4. Create and Add Complex Noise to get the Received BPSK stream 
        ComplexNoise     = (1/sqrt(2))*(sqrt(Noise_Power))*randn(1, length(QamMessage)) + ...
                      1j * (1/sqrt(2))*(sqrt(Noise_Power))*randn(1, length(QamMessage));
        Qam_Message_RX   = Qam_Message_AfterChannel + ComplexNoise;
        
        % Calculate the Received SNR
        Signal_Power                      = var(Qam_Message_RX);
        SNR                               = 10*log10(Signal_Power/Noise_Power);
        SNRdB_Array(DecisionIndex, Run)   = SNR;

        %% 5. Equalize the Received BPSK Message
        Qam_Message_Equalized    = Qam_Message_RX ./ ChannelResponse;


        %% 6. The Bit Decision
        Softbits = zeros(1, length(Qam_Message_Equalized));

        for i = 1:length(Qam_Message_Equalized)
            y = real(Qam_Message_Equalized(1,i));
    
            switch(DecisionMethod)  
                case 'Hard'
                    if(y > 0); Softbits(1,i) =  1;
                    else;      Softbits(1,i) = -1; end
             
                case 'Soft'
                    Softbits(1,i) = y;
        
                case 'LLR'
                    h = ChannelResponse(1, i);
                    Softbits(1,i) = 4 * h * conj(h) * y / Noise_Power;
            end
        end   

        %% 7. The Viterbi Decoding Process
        TracebackLength   = 100;
        BitOutput         = vitdec(Softbits, trellis, TracebackLength , 'term', 'unquant');

        Errors            = sum(xor(BitOutput, BitMessage));
        BitErrorRate      = Errors / length(BitMessage);
        BER_Array(DecisionIndex, Run) = BitErrorRate; 
    end
end

figure(1);
semilogy(SNRdB_Array(1, :), BER_Array(1, :), 'k'); grid on; hold on;
semilogy(SNRdB_Array(2, :), BER_Array(2, :), 'b');
semilogy(SNRdB_Array(3, :), BER_Array(3, :), 'r');
legend('Hard', 'Soft', 'LLR');
xlabel('Bit Error Rate');
ylabel('SNR dB');
    

ChannelResponse           = randi([30, 500], 1, length(BPSK_Message))/100;
for i = 1:length(ChannelResponse)
   if(ChannelResponse(1,i) > 1); ChannelResponse(1,i) = 1; end
end

figure(2);
plot(ChannelResponse(1, 1:500)); grid on;





% ----------------------------------------------------------------------------------
% The QAM mapper detects the number of input bits and selects the LTE constellation automatically.
% The constellations used in this function are available in TS36.211 Section 7.1
function Symbol = LteQamMapper(InputBits)

% Guarantee a row vector
InputBits = InputBits(:)';

% Error checking
ConstellationIndex = length(InputBits);    % 2 - BPSK, 3 - QPSK, 4 - 16QAM, 6 - 64QAM
ProperLength       = sum(ConstellationIndex == [1, 2, 4, 6]);
assert(ProperLength == 1, 'The input bit vector is not correctly sized.');

% Look up the right symbol
switch(ConstellationIndex)
    case 0 
        if(InputBits(1,1) == 0); I = 1/sqrt(2);   Q = 1/sqrt(2); 
        else                     I = -1/sqrt(2);  Q = -1/sqrt(2); end
    case 1 
        if(InputBits(1,1) == 0 && InputBits(1,2) == 0); I =  1/sqrt(2);   Q =  1/sqrt(2); end
        if(InputBits(1,1) == 0 && InputBits(1,2) == 1); I =  1/sqrt(2);   Q = -1/sqrt(2); end
        if(InputBits(1,1) == 1 && InputBits(1,2) == 0); I = -1/sqrt(2);   Q =  1/sqrt(2); end
        if(InputBits(1,1) == 1 && InputBits(1,2) == 1); I = -1/sqrt(2);   Q = -1/sqrt(2); end
    case 4 
        if(InputBits(1,1) == 0 && InputBits(1,3) == 0); I =  1/sqrt(10);  end
        if(InputBits(1,1) == 0 && InputBits(1,3) == 1); I =  3/sqrt(10);  end
        if(InputBits(1,1) == 1 && InputBits(1,3) == 0); I = -1/sqrt(10);  end
        if(InputBits(1,1) == 1 && InputBits(1,3) == 1); I = -3/sqrt(10);  end
        
        if(InputBits(1,2) == 0 && InputBits(1,4) == 0); Q =  1/sqrt(10);  end
        if(InputBits(1,2) == 0 && InputBits(1,4) == 1); Q =  3/sqrt(10);  end
        if(InputBits(1,2) == 1 && InputBits(1,4) == 0); Q = -1/sqrt(10);  end
        if(InputBits(1,2) == 1 && InputBits(1,4) == 1); Q = -3/sqrt(10);  end 
    case 6
        if(InputBits(1,1) == 0 && InputBits(1,3) == 0 && InputBits(1,5) == 0); I =  3/sqrt(42);  end
        if(InputBits(1,1) == 0 && InputBits(1,3) == 0 && InputBits(1,5) == 1); I =  1/sqrt(42);  end
        if(InputBits(1,1) == 0 && InputBits(1,3) == 1 && InputBits(1,5) == 0); I =  5/sqrt(42);  end
        if(InputBits(1,1) == 0 && InputBits(1,3) == 1 && InputBits(1,5) == 1); I =  7/sqrt(42);  end
        if(InputBits(1,1) == 1 && InputBits(1,3) == 0 && InputBits(1,5) == 0); I = -3/sqrt(42);  end
        if(InputBits(1,1) == 1 && InputBits(1,3) == 0 && InputBits(1,5) == 1); I = -2/sqrt(42);  end
        if(InputBits(1,1) == 1 && InputBits(1,3) == 1 && InputBits(1,5) == 0); I = -5/sqrt(42);  end
        if(InputBits(1,1) == 1 && InputBits(1,3) == 1 && InputBits(1,5) == 1); I = -7/sqrt(42);  end
        
        if(InputBits(1,2) == 0 && InputBits(1,4) == 0 && InputBits(1,6) == 0); Q =  3/sqrt(42);  end
        if(InputBits(1,2) == 0 && InputBits(1,4) == 0 && InputBits(1,6) == 1); Q =  1/sqrt(42);  end
        if(InputBits(1,2) == 0 && InputBits(1,4) == 1 && InputBits(1,6) == 0); Q =  5/sqrt(42);  end
        if(InputBits(1,2) == 0 && InputBits(1,4) == 1 && InputBits(1,6) == 1); Q =  7/sqrt(42);  end
        if(InputBits(1,2) == 1 && InputBits(1,4) == 0 && InputBits(1,6) == 0); Q = -3/sqrt(42);  end
        if(InputBits(1,2) == 1 && InputBits(1,4) == 0 && InputBits(1,6) == 1); Q = -2/sqrt(42);  end
        if(InputBits(1,2) == 1 && InputBits(1,4) == 1 && InputBits(1,6) == 0); Q = -5/sqrt(42);  end
        if(InputBits(1,2) == 1 && InputBits(1,4) == 1 && InputBits(1,6) == 1); Q = -7/sqrt(42);  end        
end

Symbol = I + 1j*Q;

end



% ----------------------------------------------------------------------------------
% This Qam Decoder will produce both Hard bits and LLR Softbits
function Bits = LteQamDemapper(QamIndex, SymbolValue, HardDecoding)

Bits = -1*ones(1, QamIndex);
% Look up the right symbol
switch(ConstellationIndex)
    case 0 
        if(HardDecoding == true)
            SymbolValue = SymbolValue * exp(-1j*pi/4);
            Softbit0    = real(SymbolValue);
            if(Softbit0 > 0); Bits(1,1) = 0;
            else;             Bits(1,1) = 1; end
        else
            
            
        end

    case 1 
        if(HardDecoding == true)
            SoftBit0 = real(SymbolValue);
            SoftBit1 = imag(SymbolValue);
            if(SoftBit0 > 0); Bits(1,1) = 0;
            else;             Bits(1,1) = 1; end
            if(SoftBit1 > 0); Bits(1,2) = 0;
            else;             Bits(1,2) = 1; end
        else
            
            
            
        end
    case 4 
        if(HardDecoding == true)
            MaxIqValue = 3 / sqrt(10);
            SymbolReal = real(SymbolValue);
            SymbolImag = imag(SymbolValue);
            % Truncate I and Q values to maximum IQ value.
            if(abs(SymbolReal) > MaxIqValue); SymbolReal = sign(SymbolReal) * MaxIqValue; end
            if(abs(SymbolImag) > MaxIqValue); SymbolImag = sign(SymbolImag) * MaxIqValue; end
            % Assign noisy symbol values to constellation positions
            NormalizedReal = 2*(floor(SymbolReal*sqrt(10) / 2) + 0.5);
            NormalizedImag = 2*(floor(SymbolImag*sqrt(10) / 2) + 0.5);
            % Compute index into the following table
            switch(NormalizedReal)
                case -3; Bits(1,[1,3]) = [1, 1];
                case -1; Bits(1,[1,3]) = [1, 0];
                case  1; Bits(1,[1,3]) = [0, 1];
                case  3; Bits(1,[1,3]) = [0, 0];
                otherwise; error('Error in Lte QAM Demapper.');              
            end
            switch(NormalizedImag)
                case -3; Bits(1,[2,4]) = [1, 1];
                case -1; Bits(1,[2,4]) = [1, 0];
                case  1; Bits(1,[2,4]) = [0, 1];
                case  3; Bits(1,[2,4]) = [0, 0];
                otherwise; error('Error in Lte QAM Demapper.');              
            end
        else
            
            
            
            
        end
        
        
    case 6
        if(HardDecoding == true)
            MaxIqValue = 7 / sqrt(42);
            SymbolReal = real(SymbolValue);
            SymbolImag = imag(SymbolValue);
            % Truncate I and Q values to maximum IQ value.
            if(abs(SymbolReal) > MaxIqValue); SymbolReal = sign(SymbolReal) * MaxIqValue; end
            if(abs(SymbolImag) > MaxIqValue); SymbolImag = sign(SymbolImag) * MaxIqValue; end
            % Assign noisy symbol values to constellation positions
            NormalizedReal = 2*(floor(SymbolReal*sqrt(42) / 2) + 0.5);
            NormalizedImag = 2*(floor(SymbolImag*sqrt(42) / 2) + 0.5);
            % Compute index into the following table
            switch(NormalizedReal)
                case -7; Bits(1,[1,3,5]) = [1, 1, 1];
                case -5; Bits(1,[1,3,5]) = [1, 1, 0];
                case -3; Bits(1,[1,3,5]) = [1, 0, 0];
                case -1; Bits(1,[1,3,5]) = [1, 0, 1];
                case  1; Bits(1,[1,3,5]) = [0, 0, 1];
                case  3; Bits(1,[1,3,5]) = [0, 0, 0];
                case  5; Bits(1,[1,3,5]) = [0, 1, 0];
                case  7; Bits(1,[1,3,5]) = [0, 1, 1];
                otherwise; error('Error in Lte QAM Demapper.');              
            end
            switch(NormalizedImag)
                case -7; Bits(1,[2,4,6]) = [1, 1, 1];
                case -5; Bits(1,[2,4,6]) = [1, 1, 0];
                case -3; Bits(1,[2,4,6]) = [1, 0, 0];
                case -1; Bits(1,[2,4,6]) = [1, 0, 1];
                case  1; Bits(1,[2,4,6]) = [0, 0, 1];
                case  3; Bits(1,[2,4,6]) = [0, 0, 0];
                case  5; Bits(1,[2,4,6]) = [0, 1, 0];
                case  7; Bits(1,[2,4,6]) = [0, 1, 1];
                otherwise; error('Error in Lte QAM Demapper.');              
            end
        else
            
            
            
            
        end
        
end


end