%% Text: Digital Signal Processing in Modern Communication Systems (Edition 2)
%% Chapter 5 - Section 3:  Exploring Gaussian FSK and Gaussian MSK
%% Copyright: Andreas Schwarzinger March 2015
%%
%% Author: Andreas Schwarzinger                          Date: March 2015

clc;
clear variables;
close all;

rand('seed', 2);
%% GFSK Paramters
BT  = 0.5;  % DECT BT  = 0.5        / GSM BT  = 0.3  / Bluetooth BT = 0.5
h   = 0.5;  % Dect h = 0.5 +/- 0.05 / GSM h   = 0.5  / Bluetooth h = 0.28
            %                                                     to 0.35 

Fsample  = 16e6;           % Sample Rate
Tsample  = 1/Fsample;      % Sample Period
Fsymbol  = 1e6;            % Symbol Period
Tsymbol  = 1/Fsymbol;      % Symbol Rate

OSR = Fsample/Fsymbol;     % Must be an integer


%% Rectangular Filter Definition
RectFilter      = ones(1, OSR);

%% Gaussian Filter Definition
Sigma           = sqrt(log(2))*Tsymbol/(BT*2*pi);
n               = -2*OSR:2*OSR;     
t               = n*Tsample;     %% The sample instances for the filter
GaussianFilter  = exp(-0.5*(t/Sigma).^2);

%% Computing the impulse response, h[n], of the Total Pulse Shaping Filter  
hn_unscaled     = conv(RectFilter, GaussianFilter);
A               = (h*pi)/(2*pi*sum(hn_unscaled)*Tsample);
hn              = A*hn_unscaled;


%% Bit Source, Symbol mapping, and zero-stuffing
BitSource                 = round(rand(1,500));  % Random bit source
SymbolSource              = 2*BitSource - 1;     % Non Return to Zero Mapping
%SymbolSource              = zeros(1,500); 
%SymbolSource(1,10)        = 1;
Zero_Stuffed              = zeros(1, OSR * length(BitSource));
Zero_Stuffed(1,1:OSR:end) = SymbolSource;

%% The filtering operation (a simple convolution)
Fd  = conv(Zero_Stuffed, hn);            

%% Acumulator and IQ Generation
%Angle = zeros(1,length(Fd));
I     = zeros(1,length(Fd));
Q     = zeros(1,length(Fd));
Phase = 0;
for i = 1:length(Fd)
    Phase      = mod(Phase + 2*pi*Fd(1,i)*Tsample , 2*pi);
    I(1,i)     = cos(Phase);
    Q(1,i)     = sin(Phase);
    %Angle(1,i) = Phase * 180/pi; 
end



figure(1);
subplot(3,1,1);
stem((1:150)*Tsample, Zero_Stuffed(1,1:150), 'k'); grid on;
title('Zero-Stuffed Sample Sequence');
xlabel('Time');
subplot(3,1,2); 
plot((40:39+150)*Tsample, sign(Fd(1,40:39+150))*2.5e5, 'k', 'Linewidth', 2);  grid on;
title('MSKFrequency Deviation Fd[n]');
xlabel('Time');
set(gca, 'YTick', [-250000 0 250000]);
subplot(3,1,3);
plot((40:39+150)*Tsample, Fd(1,40:39+150), 'k', 'Linewidth', 2);  grid on;
title('GMSK Frequency Deviation Fd[n]');
xlabel('Time');
set(gca, 'YTick', [-250000 0 250000]);


% figure(2);
% plot((40:39+150)*Tsample, Angle(1,40:39+150), 'k'); title('Angle'); grid on;



%% Compute Eye diagram

Index = 64:64:length(Fd);
Length= length(Index) - 2;

figure(2);
F = [-2.5 -2.35 -2.198 2.5 2.35 2.198]*1e5;
subplot(2,2,[1 3]);
for i = 1:Length
    Range = Index(1,i):Index(1,i) + 63;
    plot((1:64)*Tsample, Fd(1, Range), 'k'); grid on; hold on;
    plot(([9.5 9.5 9.5 9.5 9.5 9.5])*Tsample, F , 'ko');
    plot(([9.5 9.5 9.5 9.5 9.5 9.5]+16)*Tsample, F , 'ko');
    plot(([9.5 9.5 9.5 9.5 9.5 9.5]+32)*Tsample, F , 'ko');
    plot(([9.5 9.5 9.5 9.5 9.5 9.5]+48)*Tsample, F , 'ko');
end
axis([1*Tsample 64*Tsample -3e5 3e5]);

title('Frequency Deviation Eye Diagram'); xlabel('Time'); ylabel('Frequency');
 
subplot(2,2,2);
for i = 1:Length
    Range  = Index(1,i):Index(1,i) + 63;
    Range2 = Index(1,i) + [1 17 33 49];
    plot((0:63)*Tsample, I(1, Range), 'k'); grid on; hold on;
    plot(([0 16 32 48])*Tsample, I(1,Range2 - 1), 'ko');
end
title('In-Phase Eye Diagram'); xlabel('Time'); 
axis([1*Tsample 64*Tsample -1 1]);
set(gca,'XTick', [0 16 32 48]*Tsample);

subplot(2,2,4);
for i = 1:Length
    Range = Index(1,i):Index(1,i) + 63;
    Range2 = Index(1,i) + [1 17 33 49];
     plot((0:63)*Tsample, Q(1, Range), 'k'); grid on; hold on;
    plot(([0 16 32 48])*Tsample, Q(1,Range2 - 1), 'ko');
end
title('Quadrature Phase Eye Diagram'); xlabel('Time');  
axis([1*Tsample 64*Tsample -1 1]);
set(gca,'XTick', [0 16 32 48]*Tsample);

%figure(3);
%subplot(2,1,1);
%plot(I); grid on; title('I');
%subplot(2,1,2);
%plot(Q); grid on; title('Q');

figure(10);
subplot(1,2,1);
for i = 1:Length
    Range = Index(1,i):Index(1,i) + 63;
    Range2 = Index(1,i) + [1 17 33 49];
    plot(I(1, Range), Q(1, Range), 'k'); grid on; hold on;
    plot(I(1, Range2-1), Q(1, Range2-1), 'ko');
end
title('GMSK IQ Constellation Positions'); xlabel('I'); ylabel('Q');
set(gca,'XTick', [-1 0 1]);
set(gca,'YTick', [-1 0 1]);
axis([-1.2 1.2 -1.2 1.2]);

subplot(1,2,2);
plot(cos(0:.1:2*pi), sin(0:.1:2*pi), 'k'); grid on; hold on;
plot([1 0 -1 0], [0 1 0 -1], 'ko');
title('MSK IQ Constellation Positions'); xlabel('I'); ylabel('Q');
set(gca,'XTick', [-1 0 1]);
set(gca,'YTick', [-1 0 1]);
axis([-1.2 1.2 -1.2 1.2]);

figure(5);
Range = 208+5*16:16:(length(I) - 64) ;
plot(I,Q,'k'); grid on; hold on; title('IQ View');
plot(I(1,Range), Q(1,Range), 'ko');



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

Pulse1 = zeros(1,128); Pulse1(1,1)  =  1;
Pulse2 = zeros(1,128); Pulse2(1,17) = -1;
Pulse3 = zeros(1,128); Pulse3(1,33) =  1;
Pulse4 = zeros(1,128); Pulse4(1,49) =  1;
Pulse5 = zeros(1,128); Pulse5(1,65) =  1;
Pulse6 = zeros(1,128); Pulse6(1,81) =  1;

Out1   = conv(Pulse1, hn);   
Out2   = conv(Pulse2, hn);
Out3   = conv(Pulse3, hn);
Out4   = conv(Pulse4, hn);
Out5   = conv(Pulse5, hn);
Out6   = conv(Pulse6, hn);

fd     = Out1 + Out2 + Out3 + Out4 +  Out5; % + Out6;

fd1    = Out1 + Out2 + Out3 + Out4 +  Out5 + Out6;
I     = zeros(1,length(fd1));
Q     = zeros(1,length(fd1));
Phase = 0;
for i = 1:length(fd1)
    Phase      = mod(Phase + 2*pi*fd1(1,i)*Tsample , 2*pi);
    I(1,i)     = cos(Phase);
    Q(1,i)     = sin(Phase);
    %Angle(1,i) = Phase * 180/pi; 
end


 

figure(11);
x  = (1/16)*(0.5:1:206.5)*1e-6;
subplot(2,2,1);
plot(x, Out1, 'k'); grid on; hold on;
plot(x, Out2, 'k'); 
plot(x, Out3, 'k');
plot(x, Out4, 'k');
plot(x, Out5, 'k');
%Offset = 40.5;
set(gca, 'XTick', 1e-6*(1:6)); 
%set(gca, 'XTick', (1/16)*1e-6*[Offset-8 Offset+8 Offset+24 Offset+40 Offset+56 Offset+72]); 
set(gca, 'YTick', [-250e3 0 250e3]); 
axis([1e-6 7e-6 -300e3 300e3]);
title('Impulse Responses for Symbol Sequence 1, -1, 1, 1');
xlabel('Time'); ylabel('Frequency');

subplot(2,2,3);
plot(x, fd, 'k'); grid on; hold on;
Offset = 40.5;
set(gca, 'XTick', 1e-6*(1:6)); 
set(gca, 'YTick', [-250e3 0 250e3]); 
axis([1e-6 7e-6 -300e3 300e3]);
title('fd[n] for Symbol Sequence 1, -1, 1, 1');
xlabel('Time'); ylabel('Frequency');

Offset = 40;
D = [1 Offset+8 Offset+24 Offset+40 Offset+56];
subplot(2,2, [2 4]);
plot(I, Q, 'k'); grid on; hold on; 
plot(I(1,D), Q(1,D), 'ko'); 
title('Decision (Constellation) Points in the IQ Plane');
axis([-1.2 1.2 -1.2 1.2]);
xlabel('I'); ylabel('Q');
set(gca, 'XTick', [-1 0 1]); 
set(gca, 'YTick', [-1 0 1]); 
