%% Text: Digital Signal Processing in Modern Communication Systems (Edition 2)
%% Chapter 5 - Section 5:  SC-FDMA
%% Copyright: Andreas Schwarzinger (Jan 2021)
%%
%% Author: Andreas Schwarzinger                                       Date: Jan, 2021

clc;
clear variables;
close all;

%% 1. The main FFT/IFFT parameters of the OFDM portion of the SC-FDMA processes 
N_OFDM                     = 128;
N_SCFDMA                   = 32;
SubcarrierStartIndex       = 48; 
OSR                        = N_OFDM / N_SCFDMA; % The Oversampling ratio
 
%% 2. Generate BPSK (Mod Order 1) or QSPK (Mod Order 2) Input Symbols 
ModOrder = 1; 
if(ModOrder == 1)
    QamInputSequence = sign(randn(1, N_SCFDMA)) * (1 + 1j);
else
    QamInputSequence = sign(randn(1, N_SCFDMA)) + 1j*sign(randn(1, N_SCFDMA));  
end
 
%% 3. Run the SC-FDMA Transmitter Section (In the cell phone)
% 3a. Compute DFT Precoding
DftOutputBuffer    =  fft(QamInputSequence, N_SCFDMA);
%     Rearrange subcarriers from aliased to monotonically increasing frequencies.
DftOutputBuffer    = [DftOutputBuffer(1, N_SCFDMA/2 + 1: N_SCFDMA), ...
                                                         DftOutputBuffer(1, 1:N_SCFDMA/2)];
 
% 3b. Map the DftOutput into the SubcarrierBufferTx
SubcarrierBufferTx = zeros(1, N_OFDM);
SubcarrierBufferTx(1, SubcarrierStartIndex + 1 : SubcarrierStartIndex + N_SCFDMA) = ...
                                                                          DftOutputBuffer;
 
% 3c. Map the SubcarrierBufferTx into the IFFT input buffer from monotonically 
%     increasing frequency format to frequency aliased format
IfftInputBuffer                          = zeros(1, N_OFDM);
IfftInputBuffer(1, 1:N_OFDM/2)           =  SubcarrierBufferTx(1, N_OFDM/2+1:N_OFDM);
IfftInputBuffer(1, N_OFDM/2 + 1: N_OFDM) =  SubcarrierBufferTx(1, 1:N_OFDM/2);
 
% 3d. Take the IFFT and let's take a look at the signal.
IfftOutputBuffer  = (N_OFDM/N_SCFDMA)*ifft(IfftInputBuffer, N_OFDM);
 
 
%% 4. This step accomplishes the traditional single carrier down conversion to baseband
%     the signal to always see a nice baseband waveform in the plot.
fc                    = (SubcarrierStartIndex - (N_OFDM/2 - N_SCFDMA/2))/N_OFDM;
n                     = 0:N_OFDM - 1; 
UpsampledQamSequence  = IfftOutputBuffer .* exp(-1j*2*pi*n*fc); 
 
figure(1);  %% Take a look at the waveform and convince yourself that it worked
plot(real(UpsampledQamSequence), 'r'); grid on; hold on;
Range = 1:OSR:length(UpsampledQamSequence);
plot(Range, real(UpsampledQamSequence(1, Range)), 'ro');
plot(imag(UpsampledQamSequence), 'b');
plot(Range, imag(UpsampledQamSequence(1, Range)), 'bo');
 
 
%% 5.  Pass the information through the Receiver
%  5a. Take the FFT
FftOutputBuffer  = (N_SCFDMA/N_OFDM)*fft(IfftOutputBuffer, N_OFDM);
 
%  5b. Map the FFT Output in to the SubcarrierBufferRx 
%      Mapping from frequency aliased to monotonically increasing format.
SubcarrierBufferRx(1, 1:N_OFDM/2)           =  FftOutputBuffer(1, N_OFDM/2+1:N_OFDM);
SubcarrierBufferRx(1, N_OFDM/2 + 1: N_OFDM) =  FftOutputBuffer(1, 1:N_OFDM/2);
 
%  5c. Extract the desired subcarrier values SubcarrierBufferRx
BufferTemp       = SubcarrierBufferRx(1, SubcarrierStartIndex + 1 : SubcarrierStartIndex + N_SCFDMA);
 
%  5d. Rearrange subcarriers back into frequency aliased arrangements
IdftInputBuffer   = zeros(1, N_SCFDMA);
IdftInputBuffer(1, 1:N_SCFDMA/2)             = BufferTemp(1, N_SCFDMA/2 + 1: N_SCFDMA);
IdftInputBuffer(1, N_SCFDMA/2 + 1: N_SCFDMA) = BufferTemp(1, 1:N_SCFDMA/2);
QamOutputSequence = ifft(IdftInputBuffer, N_SCFDMA);
 
Error           = QamInputSequence - QamOutputSequence;
MeanSquareError = mean( Error .* conj(Error));   %% This must be tiny



return

%% 1. Setting up the main FFT/IFFT parameters of the OFDM portion of the SC-FDMA processes 
N_OFDM                     = 128;
N_SCFDMA                   = 32;
SubcarrierStartIndex       = 58; 
OSR                        = N_OFDM / N_SCFDMA; % The Oversampling ratio

% Tx Side - OFDM Section
IfftInputBuffer    = zeros(1, N_OFDM);
IfftOutputBuffer   = zeros(1, N_OFDM);
SubcarrierBufferTx = zeros(1, N_OFDM); 

% Rx Side - OFDM Section
FftOutputBuffer    = zeros(1, N_OFDM);
SubcarrierBufferRx = zeros(1, N_OFDM);

% Rx Side - SC-FDMA DFT precoder
IDftInputBuffer   = zeros(1, N_SCFDMA);
IDftOutputBuffer  = zeros(1, N_SCFDMA);

%% 2. Generate QPSK or BSPK Signals and pass them through the transmitter
ModOrder = 1;
if(ModOrder == 1)
    DftInputBuffer = sign(randn(1, N_SCFDMA)) * (1 + 1j);
else
    DftInputBuffer = sign(randn(1, N_SCFDMA)) + 1j*sign(randn(1, N_SCFDMA));  
end

% 2a. Compute DFT Precoding
DftOutputBuffer    =  fft(DftInputBuffer, N_SCFDMA);
%     Rearrange subcarriers in monotonically increasing frequencies.
DftOutputBuffer    = [DftOutputBuffer(1, N_SCFDMA/2 + 1: N_SCFDMA), DftOutputBuffer(1, 1:N_SCFDMA/2)];

% 2b. Map the DftOutput into the SubcarrierBufferTx
SubcarrierBufferTx(1, SubcarrierStartIndex + 1 : SubcarrierStartIndex + N_SCFDMA) = DftOutputBuffer;

% 2c. Map the SubcarrierBufferTx into the IFFT input buffer
IfftInputBuffer(1, 1:N_OFDM/2)           =  SubcarrierBufferTx(1, N_OFDM/2+1:N_OFDM);
IfftInputBuffer(1, N_OFDM/2 + 1: N_OFDM) =  SubcarrierBufferTx(1, 1:N_OFDM/2);

% 2d. Take the IFFT and let's take a look at the signal.
IfftOutputBuffer  = (N_OFDM/N_SCFDMA)*ifft(IfftInputBuffer, N_OFDM);

figure(1);
plot(real(IfftOutputBuffer), 'r'); grid on; hold on;
Range = 1:OSR:length(IfftOutputBuffer);
plot(Range, real(IfftOutputBuffer(1, Range)), 'ro');
plot(imag(IfftOutputBuffer), 'b');
plot(Range, imag(IfftOutputBuffer(1, Range)), 'bo');


%% 3.  Pass the information through the Receiver
%  3a. Take the FFT
FftOutputBuffer  = (N_SCFDMA/N_OFDM)*fft(IfftOutputBuffer, N_OFDM);

% 3b. Map the FFT Output in to the SubcarrierBufferRx 
SubcarrierBufferRx(1, 1:N_OFDM/2)           =  FftOutputBuffer(1, N_OFDM/2+1:N_OFDM);
SubcarrierBufferRx(1, N_OFDM/2 + 1: N_OFDM) =  FftOutputBuffer(1, 1:N_OFDM/2);

% 3c. Extract the desired subcarrier values SubcarrierBufferRx
BufferTemp       = SubcarrierBufferRx(1, SubcarrierStartIndex + 1 : SubcarrierStartIndex + N_SCFDMA);

% 3d. Rearrange subcarriers into aliased frequency arrangements
IdftInputBuffer  = zeros(1, N_SCFDMA);
IdftInputBuffer(1, 1:N_SCFDMA/2)             = BufferTemp(1, N_SCFDMA/2 + 1: N_SCFDMA);
IdftInputBuffer(1, N_SCFDMA/2 + 1: N_SCFDMA) = BufferTemp(1, 1:N_SCFDMA/2);
IdftOutputBuffer = ifft(IdftInputBuffer, N_SCFDMA)


