%% Text: Digital Signal Processing in Modern Communication Systems (Edition 2)
%% Chapter 5 - Section 6    Generating CRC Sequences
%% Copyright: Andreas Schwarzinger May 2021
%%
%% Author: Andreas Schwarzinger                              May 2021


function Remainder = GenerateCRC(Message, ModeString, InputPolynomial)

assert(nargin == 2 || nargin == 3, 'Not enough arguments to execute this function');

%% 0. The CRC generator can use the default LTE CRC polynomials or the one supplied
%     See TS36.212 Section 5.1.1 for details on the LTE CRC polynomials
switch lower(ModeString)
    case 'ltecrc24a'
        Polynomial = [1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 1, 0, 0, 1, 1, 0, 0, 1, 1, 1, 1, 1, 0, 1, 1];
    case 'ltecrc24b'
        Polynomial =  [1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 1, 1];
    case 'ltecrc16'
        Polynomial =  [1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1];
    case 'ltecrc8'
        Polynomial = [1, 1, 0, 0, 1, 1, 0, 1, 1];
    otherwise
        assert(nargin == 3, 'Not enough argument to execute this function');
        [PolyRows,    PolyColumns]    = size(InputPolynomial);
        assert(PolyRows == 1 || PolyColumns == 1, 'The input polynomial must be a vector');
        Polynomial = InputPolynomial(:).'; % Guarantee that Polynomial is a row vector
end

%% 1. Error checking
 [MessageRows, MessageColumns] = size(Message);

% 1a. Ensure the both the message and the polynomial are vectors.
assert(MessageRows == 1 || MessageColumns == 1, 'The input message must be a vector');

% 1b. Ensure that the MSB of the Polynomial is a one. It wouldn't make sense otherwise
assert(Polynomial(1,1) == 1, 'The first bit of the polyomial must be equal to 1.');

% 1c. Convert to row vector and ensure that both feature only bits
Message    = Message(:).';

for i = 1:length(Message)
    assert(Message(1,i) == 0 || Message(1,i) == 1,           'The Message must be composed of bits');
    if(i <= length(Polynomial))
        assert(Polynomial(1,i) == 0 || Polynomial(1,i) == 1, 'The Polynomial must be composed of bits');
    end
end


%% 2. Compute the CRC process
%     Create a temporary message by appending length(Polynomial) - 1 zeros to the original message.
TempMessage = [Message, zeros(1, length(Polynomial) - 1)];
for i = 1:length(Message) 
   Range = i:(i+length(Polynomial)-1);
   if(TempMessage(1,i) ~= 0)
       % The modulo 2 addition
       TempMessage(1, Range) = mod(TempMessage(1, Range) + Polynomial, 2);
   end
   if(sum(TempMessage) == 0); break; end 
end

% CRC remainder as a row vector
Remainder = TempMessage(1, (end - length(Polynomial) + 2):end);

% Change the remainder to a column vector if the original message was also a column vector.
if(MessageColumns == 1); Remainder = Remainder.'; end




