%% Text: Digital Signal Processing in Modern Communication Systems (Edition 2)
%% Chapter 5 - Section 6    The aim of this script is to illustrate the advantage of LLR scaling ahead of binary convolutional
%%         decoding.
%% Copyright: Andreas Schwarzinger May 2021
%%
%% Author: Andreas Schwarzinger                              May 2021

clc;
clear variables;
close all;

%% 0. Some definitions regarding the convolutional encoder
%     The specifications used here are those of the LTE PBCH BCC encoder (See 3GPP TS36.212 Sect. 5.1.3.1

DecisionMethod       = 'Hard';   % 'Soft', 'LLR', 'Hard'
DecisionMethods      = cell(3,1);
DecisionMethods{1,1} = 'Hard';
DecisionMethods{2,1} = 'Soft';
DecisionMethods{3,1} = 'LLR';

ConstraintLength = 7;  
trellis          = poly2trellis(ConstraintLength,[133,171,165]);
InitState        = 0;

Noise_Power  = 0.25;
NumRuns      = 25;
SNRdB_Array  = zeros(3, NumRuns); 
BER_Array    = zeros(3, NumRuns);

for DecisionIndex = 1:length(DecisionMethods)
    DecisionMethod = DecisionMethods{DecisionIndex, 1};
    Noise_Power  = 0.1;
    for Run = 1:NumRuns
        Noise_Power = Noise_Power*1.2;
%    SNR_dB           = 0;
%    SNR_Linear       = 10^(SNR_dB/10);  
%    Signal_Power     = 1;               % As the BPSK symbols are equal to +/-1   +  j0
%    Noise_Power      = Signal_Power/SNR_Linear;
    

        %% 1. Define the input bit sequence
        NumberOfInfoBits = 50000;
        InfoBits         = randi([0, 1], 1, NumberOfInfoBits);
        NumberOfPadBits  = ConstraintLength - 1;
        PadBits          = zeros(1, NumberOfPadBits);
        BitMessage       = [InfoBits, PadBits];


        %% 2. Encode the BitMessage
        %     The encoding step            
        [EncodedBits]    = convenc(BitMessage, trellis, InitState);

        %% 2a  Convert to inverted bipolar --> Map bit value = 0 to +1.0,   and a bit value = 1 to -1.0
        BPSK_Message     = -(EncodedBits*2 - 1); 


        %% 3. Distort via Channel Response and Noise
        ChannelResponse           = randi([30, 500], 1, length(BPSK_Message))/100;
        for i = 1:length(ChannelResponse)
            if(ChannelResponse(1,i) > 1); ChannelResponse(1,i) = 1; end
        end
        BPSK_Message_AfterChannel = BPSK_Message.*ChannelResponse;
    

        %% 4. Create and Add Complex Noise to get the Received BPSK stream 
        ComplexNoise     = (1/sqrt(2))*(sqrt(Noise_Power))*randn(1, length(BPSK_Message)) + ...
                      1j * (1/sqrt(2))*(sqrt(Noise_Power))*randn(1, length(BPSK_Message));
        BPSK_Message_RX  = BPSK_Message_AfterChannel + ComplexNoise;
        
        % Calculate the Received SNR
        Signal_Power                      = var(BPSK_Message_AfterChannel);
        SNR                               = 10*log10(Signal_Power/Noise_Power);
        SNRdB_Array(DecisionIndex, Run)   = SNR;

        %% 5. Equalize the Received BPSK Message
        BPSK_Message_Equalized    = BPSK_Message_RX ./ ChannelResponse;


        %% 6. The Bit Decision
        Softbits = zeros(1, length(BPSK_Message_Equalized));

        for i = 1:length(BPSK_Message_Equalized)
            y = real(BPSK_Message_Equalized(1,i));
    
            switch(DecisionMethod)  
                case 'Hard'
                    if(y > 0); Softbits(1,i) =  1;
                    else;      Softbits(1,i) = -1; end
             
                case 'Soft'
                    Softbits(1,i) = y;
        
                case 'LLR'
                    h = ChannelResponse(1, i);
                    Softbits(1,i) = 4 * h * conj(h) * y / Noise_Power;
            end
        end   

        %% 7. The Viterbi Decoding Process
        TracebackLength   = 100;
        BitOutput         = vitdec(Softbits, trellis, TracebackLength , 'term', 'unquant');

        Errors            = sum(xor(BitOutput, BitMessage));
        BitErrorRate      = Errors / length(BitMessage);
        BER_Array(DecisionIndex, Run) = BitErrorRate; 
    end
end

figure(1);
semilogy(SNRdB_Array(1, :), BER_Array(1, :), 'k'); grid on; hold on;
semilogy(SNRdB_Array(2, :), BER_Array(2, :), 'b');
semilogy(SNRdB_Array(3, :), BER_Array(3, :), 'r');
legend('Hard', 'Soft', 'LLR');
ylabel('Bit Error Rate');
xlabel('SNR dB');
    

ChannelResponse           = randi([30, 500], 1, length(BPSK_Message))/100;
for i = 1:length(ChannelResponse)
   if(ChannelResponse(1,i) > 1); ChannelResponse(1,i) = 1; end
end

figure(2);
plot(ChannelResponse(1, 1:500)); grid on;