%% Text: Digital Signal Processing in Modern Communication Systems (Edition 2)
%% Chapter 5 - Section 6    The following script will exercise both the encoding and decoding steps for the 802.11n LDPC
%%                          implementations
%% Copyright: Andreas Schwarzinger May 2021
%%
%% Author: Andreas Schwarzinger                              May 2021

clc;
clear variables;

%% 1. Define the Prototype Matrix (see 802.11n specification - reference [1])
ProtoType = [  0  -1  -1  -1   0   0  -1  -1   0  -1  -1   0   1   0  ...
                                              -1  -1  -1  -1  -1  -1  -1  -1  -1  -1; ...
              22   0  -1  -1  17  -1   0   0  12  -1  -1  -1  -1   0  ...
                                               0  -1  -1  -1  -1  -1  -1  -1  -1  -1; ...
               6  -1   0  -1  10  -1  -1  -1  24  -1   0  -1  -1  -1   ...
                                               0   0  -1  -1  -1  -1  -1  -1  -1  -1; ...
               2  -1  -1   0  20  -1  -1  -1  25   0  -1  -1  -1  -1   ...
                                              -1   0   0  -1  -1  -1  -1  -1  -1  -1; ...
              23  -1  -1  -1   3  -1  -1  -1   0  -1   9  11  -1  -1   ...
                                              -1  -1   0   0  -1  -1  -1  -1  -1  -1; ...
              24  -1  23   1  17  -1   3  -1  10  -1  -1  -1  -1  -1  ...
                                              -1  -1  -1   0   0  -1  -1  -1  -1  -1; ...
              25  -1  -1  -1   8  -1  -1  -1   7  18  -1  -1   0  -1  ...
                                              -1  -1  -1  -1   0   0  -1  -1  -1  -1; ...
              13  24  -1  -1   0  -1   8  -1   6  -1  -1  -1  -1  -1  ...
                                              -1  -1  -1  -1  -1   0   0  -1  -1  -1; ...
               7  20  -1  16  22  10  -1  -1  23  -1  -1  -1  -1  -1  ...
                                              -1  -1  -1  -1  -1  -1   0   0  -1  -1; ...
              11  -1  -1  -1  19  -1  -1  -1  13  -1   3  17  -1  -1  ...
                                              -1  -1  -1  -1  -1  -1  -1   0   0  -1; ...
              25  -1   8  -1  23  18  -1  14   9  -1  -1  -1  -1  -1  ...
                                              -1  -1  -1  -1  -1  -1  -1  -1   0   0; ...
               3  -1  -1  -1  16  -1  -1   2  25   5  -1  -1   1  -1  ...
                                              -1  -1  -1  -1  -1  -1  -1  -1  -1   0];

SubmatrixDimension = 27;
H = CreateParityCheckMatrix(ProtoType, SubmatrixDimension);
N = size(H,2);   % Number of encoded bits = number columns in matrix    
L = size(H,1);   % Number of parity bits (check nodes) = number of rows in matrix      
K = N - L;       % Number of message bits

%% 2. Compute the Gaussian elimination to force H into desired format.
H_Modified = UseGaussianElimination(H);
P          = H_Modified(1:L, 1:K);
G          = [eye(K), P'];         % The generator matrix G

%     To ensure that the generator matrix is valid perform the following check.
Error= sum(sum(mod(G*H', 2), 1), 2);
assert(Error == 0, 'An error occurred while generating G.');


%% 3. Generate the encoded bits based on the following Input bit vector
InputBits             = zeros(1, K);
InputBits(1, 2:2:end) = ones(1, K/2);              % InputBits = [1,  0, 1,  0, 1,  0, ...., 1,  0] 
OutputBits            = mod(InputBits*G, 2);
TxBeliefs             = (2*OutputBits - 1);        % Convert from 0/1 to -1/+1 

% Generate and add noise to the output bits
rng(0);
SnrDb          = 2; 
SnrLinear      = 10^(SnrDb/10);
NoiseVariance  = 1/SnrLinear;
NoiseVector    = sqrt(NoiseVariance)*randn(size(TxBeliefs));
RxBeliefs      = (TxBeliefs + NoiseVector);


%% 4. Let's do the Min Sum LDPC decoding
NumberOfIterations  = 10;
IntrinsicBeliefs    = RxBeliefs;
E                   = zeros(size(H));
for Iteration = 1:NumberOfIterations
    % Step 1 and Step 4
    M = IntrinsicBeliefs .* ones(size(H));   %% Populate the M Matrix
    %  Zero out positions that are not interesting to us. This also helps
    %  when we add values column wise to find the total extrinsic belief
    M = M .* H;       
    M = M - E;        % During the first pass, E  = 0 (Step 1)
                      % During later passes,   E ~= 0 (Step 4)

    % Step 2 and Step 5:
    for CheckNodeIndex = 1:L
        % Just grab the current row in the parity check matrix i.e. [0 1 0 0 1 1 0 0 ...]
        CheckNodeConnectionsVector = H(CheckNodeIndex, :);
        % Find the indices of those connections                i.e. [2, 5, 6, ...]
        CheckNodeConnectionIndices = find(CheckNodeConnectionsVector);
        % Fetch the extrinsic beliefs at that row and send to SISO_SPC_Decoder
        r                          = M(CheckNodeIndex, CheckNodeConnectionIndices);
        l                          = SISO_SPC_Decoder(r);
        % This matrix remembers the intrinsic beliefs that we need to subtract
        E(CheckNodeIndex, CheckNodeConnectionIndices) = l;
    end
    % Step 4:
    % Find the sum of the extrinsic beliefs for a particular bit node.
    TotalExtrinsicBeliefs = sum(E, 1); 
    % Add that sum to the intrinsic belief (original received bit beliefs) to get the new updated
    % intrinsic belief.
    IntrinsicBeliefs      = SISO_Repetition_Decoder(RxBeliefs, TotalExtrinsicBeliefs);
end 

disp(['Rx Beliefs:       ' num2str(RxBeliefs)]);
disp(['Final Beliefs:    ' num2str(IntrinsicBeliefs)]);
% Map received beliefs back to bits.   Minus sign --> 0   and Plus sign --> 1
RxBitEstimates    = 0.5*(sign(IntrinsicBeliefs(1, 1:K))+1);
disp(['Rx Bit Estimates: ' num2str(RxBitEstimates)]);
FinalMessageBitErrors  = sum(mod(RxBitEstimates + InputBits , 2));
disp(['Remaining Errors: ' num2str(FinalMessageBitErrors)]);





% r - Is the vector of intrinsic beliefs at the bit nodes
% l - Is the vector of extrinsic belief for the single parity check node at index x
function [l, l2] = SISO_SPC_Decoder(r)
    l  = zeros(size(r));
    for x = 1:length(r)
        r_other      = r;                      
        r_other(x)   = [];                     % Remove rx from vector r and store in r_other
        Sign_Other   = -sign(prod(-r_other));  % Compute sign of other beliefs
        Mag_Other    = min(abs(r_other));      % Compute the minimum abs value of other beliefs
        l(x)         = Sign_Other * Mag_Other; % Build the extrinsic belief vector
    end   
end




% rx  -  The current instrinsic belief regarding the received bit
% lx  -  A vector of extrinsic beliefs provided by each check node to which rx connects.
function UpdatedIntrinsicBelief = SISO_Repetition_Decoder(rx, lx)
    UpdatedIntrinsicBelief = rx + lx;
end


% Produce the full parity matrix based on the prototype matrix defined in the specification
% M         - The prototype matrix
% Dimension - The dimensions of each submatrix
function H = CreateParityCheckMatrix(M, Dimension)
    SizeH = size(M)*Dimension;       % The size if rows and columns of the parity check matrix
    H     = zeros(SizeH);            % The paritiy check matrix initialized to all zeros
    for row = 1:size(M, 1)
        for col = 1:size(M, 2)
            % Extract the cyclic shift from the prototype matrix 
            CyclicShift   = M(row, col); 
            % The cyclically shifted identity matrix
            I_Shifted     = circshift( eye(Dimension), CyclicShift, 2); 
            % Figure out the row and column ranges inside H where we would insert I_Shifted
            RowRange      = (1:Dimension) + (row - 1)*Dimension; 
            ColRange      = (1:Dimension) + (col - 1)*Dimension; 
            % Actually insert it only if the cyclic shift is positive. 
            % A negative shift in the prototype matrix indicates that this submatrix is all zeros.
            if(CyclicShift >= 0)
                H(RowRange, ColRange) = I_Shifted;
            end
        end
    end
end



% H     -  The parity check matrix
% H_New -  The new parity matrix featuring an LxL identity matrix at the right side
function H_New = UseGaussianElimination(H)
    H_New = H;  
    [rows, columns] = size(H_New);
    N               = columns;  % The number of encoded bits
    L               = rows;     % The number of parity bits
    K               = N - L;    % The number of message bits

    % Check to see whether H is already in the desired form (with I on the right side).
    % Unfortunatly it likely won't be.
    I = H_New(1:L, (K+1):N);
    if(sum(sum(mod(I + eye(L), 2), 1), 2) == 0)   % No need to run Gauss-Jordan Elimination
        return;
    end
    
    % ------------------------------------------------------------------
    % Step 1: The forward elimination step        
    for column = K+1 : N  
        % Forward elimination is happening in this column
        % The row we are currently starting from.
        RowOfInterest = column - K;  % Increments as follows: 1, 2, 3 .... L
        
        % Determine whether we need to pivot or not.
        if(H_New(RowOfInterest, column) ~= 1) % then we must pivot
            % The situation here is a bit different than in regular matrices. 
            % Here, we can pivot row and columns
            % In the current column, check to see wether there are any '1's in the rows below
            % the row of interest. If there are no rows left, then we need to pivot the column
            RemainingRowsBelow = L - RowOfInterest;
            PivotSuccessful = false;
            if(RemainingRowsBelow > 0)
                % We attempt to pivot rows
                for row = RowOfInterest+1:L
                    if(H_New(row, column) == 1)
                        H_New = AddRows(H_New, row, RowOfInterest); 
                        PivotSuccessful = true;
                        break; 
                    end  
                end
            end
            
            if(RemainingRowsBelow == 0 || PivotSuccessful == false)
               for ColIndex = 1:N
                   if(H_New(RowOfInterest, ColIndex) == 1)
                       H_New = SwitchColumns(H_New, ColIndex, column);
                       PivotSuccessful = true;
                       break;
                   end
                end                
            end  
            assert(PivotSuccessful, 'The Pivot operation failed.');
        end
        
        % Forward elimination
        for row = RowOfInterest+1:L
             if(H_New(row, column) == 1); H_New = AddRows(H_New, RowOfInterest, row); end  
        end
    end
    
    % ------------------------------------------------------------------
    % Step 2: The backward elimination step 
    for column = N : - 1: K + 2
        RowOfInterest = column - K;   % Decrements as follows: L, L-1, L-2, ... 2
        % If any of the rows above feature a 1 in this column, then get rid of it by adding.
        for row = 1:RowOfInterest - 1
            if(H_New(row, column) == 1); H_New = AddRows(H_New, RowOfInterest, row); end  
        end
    end
end


% The following functions adds the source row of InputM to the target row of InputM.
function OutputM = AddRows(InputM, SourceRow, TargetRow)
    OutputM               = InputM;
    OutputM(TargetRow, :) = mod(OutputM(TargetRow, :) + OutputM(SourceRow, :), 2);
end



