%% Text: Digital Signal Processing in Modern Communication Systems (Edition 2)
%% Chapter 5 - Section 6    This is the Successive Cancellation decoder as presented in Chapter 5.6.5
%% Copyright: Andreas Schwarzinger Oct 2021
%%
%% Author: Andreas Schwarzinger                              Oct 2021

function Output = PolarDecoderBook(Y, FrozenPositions0Based)

N              = length(Y);
NumberOfStages = log10(N)/log10(2);  % Number of stages in structure
if(rem(NumberOfStages, 1) ~= 0); error('N must be equal to 2^integer'); end
 
% Define state of the nodes:
% Nothing Done = 0, MinSum Decoding Done = 1, Repetition Decoding Done = 2
States = cell(NumberOfStages, 1);
for stage = 1:NumberOfStages
   States{stage, 1} = zeros(1, 2^(NumberOfStages - stage) ); 
end
 
% Define the Belief matrix and place the received beliefs into the right most column
L = zeros(N, NumberOfStages + 1);
L(:, NumberOfStages + 1)          = Y(:);
 
% Define the bit matrix and place the known frozen values into the left most column
B = -10*ones(N, NumberOfStages + 1);
B(FrozenPositions0Based+1, 1)     = zeros(length(FrozenPositions0Based), 1);
 
CurrentNode  = 0;                 % The start node
CurrentStage = NumberOfStages;    % The start state
 
while true
    % Check the state of ths current node and see what needs to be done  
    NodeState = States{CurrentStage, 1}(1, CurrentNode + 1);
    % If 0 -> We need to run the MinSum Decoding    (Moving right to left)
    % If 1 -> We need to run the repetition decoder (Moving right to left)
    % If 2 -> Re-encode Bit information             (Moving left to right)
    % If 3 -> The node is done. Current configuration should never be reached.
    % Get the ports for the current node and stage
    [~, CurrentUpperIndices, CurrentLowerIndices] = ...
                                      GetPortIndices(CurrentNode, CurrentStage);
    switch(NodeState)
        case 0   % We will do the MinSum decoding operation 
            UpperBeliefs                 = L(CurrentUpperIndices + 1, CurrentStage+1);
            LowerBeliefs                 = L(CurrentLowerIndices + 1, CurrentStage+1);
                                  
            if(CurrentStage == 1)   
                % If we are in stage 1:
                % 1. Run the MinSumDecoder and deposit result in L
                OutputBelief            = MinSumDecoder(UpperBeliefs, LowerBeliefs);
                L(CurrentUpperIndices + 1, CurrentStage) = OutputBelief;
                % 2. Make a hard decision unless we have reached a frozen bit.
                if(B(CurrentUpperIndices + 1,1)~=0)  % Hard decision unless frozen bit
                    B(CurrentUpperIndices + 1, CurrentStage) = ... 
                                               0.5*(sign(OutputBelief + 1e-10) + 1);
                end
                U            = B(CurrentUpperIndices + 1, 1);
                % 3. Run the repetition decoder.
                OutputBelief = RepetitionDecoder(UpperBeliefs, LowerBeliefs, U);
                L(CurrentLowerIndices + 1, CurrentStage) = OutputBelief;
                % 4. Make a hard decision unless we have reached a frozen bit.
                if(B(CurrentLowerIndices + 1, 1)~=0)  % Hard decision unless frozen bit
                    B(CurrentLowerIndices + 1, CurrentStage) = ... 
                                             0.5*(sign(OutputBelief + 1e-10) + 1);
                end
                % 5. Run the re-encoding step.
                B(CurrentUpperIndices + 1, 2) = mod(B(CurrentUpperIndices + 1, 1)+ ...
                                                    B(CurrentLowerIndices + 1, 1), 2);
                B(CurrentLowerIndices + 1, 2) = B(CurrentLowerIndices + 1, 1);
                % MinSum and Repetition done
                States{CurrentStage, 1}(1, CurrentNode + 1) = 2;                
            else
                % If we are in stage > 1
                % 1. Run the MinSumDecoder and deposit result in L
                OutputBeliefs           = MinSumDecoder(UpperBeliefs, LowerBeliefs);
                L(CurrentUpperIndices + 1, CurrentStage) = OutputBeliefs;
                States{CurrentStage, 1}(1, CurrentNode + 1) = 1; % MinSum decoding done
            end
        case 1   % We will do the Repetition decoding operation for stages > 1
            UpperBeliefs   = L(CurrentUpperIndices + 1,  CurrentStage + 1);
            LowerBeliefs   = L(CurrentLowerIndices + 1,  CurrentStage + 1);
            U              = B(CurrentUpperIndices  + 1, CurrentStage);
            OutputBeliefs  = RepetitionDecoder(UpperBeliefs, LowerBeliefs, U);
            L(CurrentLowerIndices + 1, CurrentStage) = OutputBeliefs;
            States{CurrentStage, 1}(1, CurrentNode + 1) = 2; % Repetition decoding done
        case 2
            % Now that we finished the repetition decoding operation we will now do the
            % re-encoding
            UpperBits      = B(CurrentUpperIndices + 1, CurrentStage);
            LowerBits      = B(CurrentLowerIndices + 1, CurrentStage);
            B(CurrentUpperIndices + 1, CurrentStage+ 1) = mod(UpperBits+LowerBits, 2);
            B(CurrentLowerIndices + 1, CurrentStage+ 1) = LowerBits;
            if(CurrentStage == NumberOfStages); break; end
        otherwise
            error('The node is done with all computations.');
    end

    % Let's the current node and stage
    % If we are already in stage 1, then we must move from right to left
    MovingLeft = false;
    if((NodeState == 0 || NodeState == 1) && CurrentStage ~= 1); MovingLeft = true; end
    [CurrentStage, CurrentNode] = GetNextNode(CurrentNode   ...
                                            , CurrentStage  ...
                                            , MovingLeft    ...
                                            , NodeState);
end
Output = B(:,1);
end


% ------------------------------------------------------------------------------------
% Function: RepetitionDecoder()
% ------------------------------------------------------------------------------------
function OutputBelief = MinSumDecoder(Y1, Y2)
OutputBelief          = -sign(Y1).*sign(Y2).*min(abs(Y1), abs(Y2));
end

% ------------------------------------------------------------------------------------
% Function: RepetitionDecoder()
% ------------------------------------------------------------------------------------
function OutputBelief = RepetitionDecoder(Y1, Y2, U)
OutputBelief = zeros(length(U), 1);
for Index = 1:length(U)
    if(U(Index, 1) <= 0); OutputBelief(Index, 1) = Y2(Index,1) + Y1(Index, 1);  
    else                ; OutputBelief(Index, 1) = Y2(Index,1) - Y1(Index, 1); end
end
end



% ------------------------------------------------------------------------------------
% Function: GetNextNode()
% ------------------------------------------------------------------------------------
function [NextStage, NextNode] = GetNextNode(CurrentNode   ...
                                           , CurrentStage  ...
                                           , MovingLeft    ...
                                           , NodeState)
% If we are moving left, which is the decoding direction producing beliefs
if(MovingLeft == true)   
    NextStage   = CurrentStage - 1;      % As we are moving left
    NextNode    = CurrentNode * 2;       % If MinSum     decoder (Upper node)
    if(NodeState == 1)                   % If repetition decoder (Lower node)
        NextNode = NextNode + 1;
    end
% Else we are moving right, which is the re-encoding direction producing 0/1 bits.
else                      
    NextStage   = CurrentStage + 1;            % As we are moving right
    NextNode    = floor(CurrentNode / 2);
end
end


% ------------------------------------------------------------------------------------
% Function: GetPortIndices()
% ------------------------------------------------------------------------------------
function [AllIndices, UpperIndices, LowerIndices] = ...
                                             GetPortIndices(CurrentNode, CurrentStage)
NumberOfPorts = 2^CurrentStage;
StartPort     = CurrentNode * NumberOfPorts;
AllIndices    = StartPort : StartPort + NumberOfPorts - 1;      % -1 because this is a 
UpperIndices  = AllIndices(1, 1:NumberOfPorts/2);               % 0 based index
LowerIndices  = AllIndices(1, NumberOfPorts/2 + 1 :NumberOfPorts);  
end

