%% Text: Digital Signal Processing in Modern Communication Systems (Edition 2)
%% Chapter 5 - Section 6    Testing the GaussJordan Elimination to get the Generator Matrix from
%%                          a parity check matrix.
%% Copyright: Andreas Schwarzinger May 2021
%%
%% Author: Andreas Schwarzinger                              May 2021

clc;
H = [1 1 1 1 0 0; ...
     1 0 0 1 1 0; ...
     0 0 1 1 0 1];

H_New = GaussJordanElimination(H)


% H     -  The parity check matrix
% H_New -  The new parity matrix featuring an LxL identity matrix at the right side
function H_New = GaussJordanElimination(H)
    H_New = H;  
    [rows, columns] = size(H_New);
    N               = columns;  % The number of encoded bits
    L               = rows;     % The number of parity bits
    K               = N - L;    % The number of message bits
    
    % ------------------------------------------------------------------
    % Step 1: The forward elimination step        
    for column = K+1 : N  
        % [RowOfInterest, column] are the coordinates of the diagonal of the LxL 
        % square matrix all the way to the right of H_New. The point of the exercise
        % is to place 1's in these positions. If we don't find a 1 there, we need 
        % to look to see whether one of the rows below has a one in this column 
        % and then pivot.
        RowOfInterest = column - K;  % Increments as follows: 1, 2, 3 .... L
        
        % 1a. The pivoting step (if we need to pivot)
        % Determine whether we need to pivot or not.
        if(H_New(RowOfInterest, column) ~= 1) 
            % We will now look at the rows below until we find a 1 in this column.
            % Rather than switching the current row with the one containing the 1,
            % we add the row below to the current one that is missing the 1. This is fine
            % as we are doing modulo 2 addition. We could have switched them as well.
            RemainingRowsBelow = L - RowOfInterest;
            PivotSuccessful = false;
            if(RemainingRowsBelow > 0)
                % We attempt to pivot rows
                for row = RowOfInterest+1:L
                    if(H_New(row, column) == 1)
                        H_New = AddRows(H_New, row, RowOfInterest); 
                        PivotSuccessful = true;
                        break; 
                    end  
                end
            end
            assert(PivotSuccessful, 'The Pivot operation failed. Go make a better H.');
        end
        
        % At this point, we have ensured that the diagonal of the square matrix on
        % the right of H_New features ones everywhere.
        % 1b. Forward elimination
        for row = RowOfInterest+1:L
             if(H_New(row, column) == 1); H_New = AddRows(H_New, RowOfInterest, row); end  
        end
    end
    
    % ------------------------------------------------------------------
    % Step 2: The backward elimination step 
    for column = N : - 1: K + 2
        RowOfInterest = column - K;   % Decrements as follows: L, L-1, L-2, ... 2
        % If any of the rows above feature a 1 in this column, then get rid of it by adding.
        for row = 1:RowOfInterest - 1
            if(H_New(row, column) == 1); H_New = AddRows(H_New, RowOfInterest, row); end  
        end
    end
end

% The following functions adds the source row of InputM to the target row of InputM.
function OutputM = AddRows(InputM, SourceRow, TargetRow)
    OutputM               = InputM;
    OutputM(TargetRow, :) = mod(OutputM(TargetRow, :) + OutputM(SourceRow, :), 2);
end