%% Text: Digital Signal Processing in Modern Communication Systems (Edition 2)
%% Chapter 7 - Section 1.3.5: This script computes the channel and RS-CINR estimate of a raw frequency response for the
%%        following scenarios:
%%        LTE (1.4MHz / Antenna port 0 / 1 subframe)
%%        C-V2X:   (Over a Single PSCCH Resource - 1 subframe / 36 Subcarriers)
%%        802.11a: (Over 52 Subcarriers)
%% Copyright: Andreas Schwarzinger May 2021
 

clc;
clear variables;
close all;

MeanSquare     = @(x) ( sum(sum(x.*conj((x)),1),2) / length(x(:)) );

%% Step 1. Set the reference signal coordinates and channel model given the standard
Standard = 'WLAN';   % LTE / V2X / WLAN
if     (strcmp(Standard, 'LTE') == 1)
    SubcarrierSpacing = 15e3;
    OfdmSymbolPeriod  = 0.001/12; % Extended Cyclic Prefix (12 symbols per subframe)
    Frequencies  = [-36:6:-1,  1:6:36, ...                     % Symbol l = 0
                    -33:6:-1,  4:6:36, ...                     % Symbol l = 3
                    -36:6:-1,  1:6:36, ...                     % Symbol l = 6
                    -33:6:-1,  4:6:36] * SubcarrierSpacing;    % Symbol l = 9
    Times        = [zeros(1,12), 3*ones(1,12), 6*ones(1,12), 9*ones(1,12)]*OfdmSymbolPeriod;
    DelayRange   = [-1e-6, 4e-6];                              % Channel Model
    DopplerRange = [-300 300];                                 % Channel Model 
elseif (strcmp(Standard, 'V2X') == 1)
    SubcarrierSpacing = 15e3;
    OfdmSymbolPeriod  = 0.001/14; % Normal Cyclic Prefix (14 symbols per subframe)
    Frequencies = [0:23, ...                                  % Symbol l = 2
                   0:23, ...                                  % Symbol l = 5
                   0:23, ...                                  % Symbol l = 8
                   0:23]* SubcarrierSpacing;                  % Symbol l = 11
    Times       = [2*ones(1,24), 5*ones(1,24), 8*ones(1,24), 11*ones(1,24)]*OfdmSymbolPeriod;
    DelayRange   = [-0.4e-6, 2e-6];                           % Channel Model
    DopplerRange = [-750 750];                                % Channel Model 
elseif (strcmp(Standard, 'WLAN') == 1)
    SubcarrierSpacing = 312.5e3; % Hz
    OfdmSymbolPeriod  = 4e-6;    % seconds   
    Frequencies = [-26:1:-1,  1:1:26]* SubcarrierSpacing; 
    Times       = zeros(1,52)*OfdmSymbolPeriod;
    DelayRange   = [-50e-9 250e-9];                           % Channel Model
    DopplerRange = [-10 10];                                   % Channel Model 
else
    assert(false, 'The selection is unsupported.');
end

%% Step 2. Compute the Frequency response for a large set of channel conditions 
DelayStep         = (DelayRange(1,2)   - DelayRange(1,1))   / 100;
DopplerStep       = (DopplerRange(1,2) - DopplerRange(1,1)) / 50;
DelayVector       = DelayRange(1,1):DelayStep:DelayRange(1,2);        
DopplerVector     = DopplerRange(1,1):DopplerStep:DopplerRange(1,2);  
NumChannels       = length(DelayVector)*length(DopplerVector); % The number of channel conditions
NumCoordinates    = length(Frequencies);                       % The number reference signals

rng(3);
FreqResponseMatrix  = zeros(NumChannels, NumCoordinates);   
NumChannelCondition = 0;
for A = 1:length(DelayVector)
    Delay = DelayVector(1,A);                    % Grab the delay from source vector
    for B = 1:length(DopplerVector)
        NumChannelCondition = NumChannelCondition + 1;
        Doppler = DopplerVector(1,B);            % Grab the doppler from the source vector
        % The time delay will cause the frequency response to rotate over frequency.
        % The doppler will cause the frequency response to rotate over time.
        % These rotations need to start at a random angle in order for the NumCoordinates 
        % dimensional space to be covered properly. If we don't do this, the eigenvalues and 
        % eigenvectors produced by the PCA will not work properly.
        StartAngle           = rand(1,1)*2*pi;    
        for Position = 1:NumCoordinates
           Frequency         = Frequencies(1, Position);
           Time              = Times(1, Position);
           % The Complex scalar of the path
           C_Scalar          = exp(1j*StartAngle);       
           % Impact on the frequency response at this reference signal due to time delay.
           AngleDueToDelay   = -2*pi*Frequency * Delay;  
           C_Delay           = exp(1j*AngleDueToDelay);
           % Impact on the frequency response at this reference signal due to Doppler.
           AngleDueToDoppler = 2*pi*Doppler* Time; 
           C_Doppler         = exp(1j*AngleDueToDoppler);
           % The final frequency response at the current reference signal coordinate.
           C_FreqResponse    = C_Scalar * C_Delay * C_Doppler;
           % Hold the total frequency response of each channel condition in the rows of the matrix below.
           FreqResponseMatrix(NumChannelCondition, Position) = C_FreqResponse; 
        end
    end
end


%% Step 3. Compute the Covariance Matrix and the Eigen value decomposition.
CovarianceM = cov(conj(FreqResponseMatrix)); % NumCoordinates by NumCoordinates matrix
[V, D]      = eig(CovarianceM);             % The eigenvalue decomposition
% Grab the Eigenvectors, which are in the columns of V
EigenVectors = V;
% Grab the Eigenvalues, which are in the diagonal of D
EigenValues = diag(D);

figure(1);
stem(EigenValues, 'k'); grid on; 
title('EigenValues  Data Extent Along each Principle Component');



%% Step 4. Generate a typical frequency response and force a particular CINR.
                %    Cp      Delay(sec)   Doppler
ChannelCondition = [  1,     200e-9,       0;  ... % Path 0 (The strongest path)
                  -0.7j,     -20e-9,       0;  ... % Path 1 (The earliest path)
             0.3 - 0.7j,      80e-9,       0;  ... % Path 2
                     1j,     120e-9,       0;  ... % Path 3  
                    0.4,     200e-9,       0;  ... % Path 4  
             0.1 - 0.1j,     250e-9,       0];     % Path 5 (The Latest path)

%   Compute the frequency response (See code from Section 7.1.3.1)
FreqResponse      = ComputeFrequencyResponse(ChannelCondition(:,:), Frequencies, Times);

%   Force a particular reference signal CINR (carrier to interference and noise ratio)
FreqResponsePower = MeanSquare(FreqResponse);
RsCinrdB          = 5;                         % dB
RsCinrLinear      = 10^(RsCinrdB/10);
NoisePower        = FreqResponsePower/RsCinrLinear;
GaussianNoise     = 0.7071*sqrt(NoisePower)*(randn(1, length(FreqResponse)) + ...
                                        + 1j*randn(1, length(FreqResponse)) );

% The raw frequency response as a column vector
RawFreqResponse   = (FreqResponse + GaussianNoise).'; 


%% Step 5. Generate a typical frequency response and force a particular CINR
% Compute seperator1 and seperator2
MaxEigenValue       = max(EigenValues);
SeperatorThreshold1 = 0.02  * MaxEigenValue;                  % Good for RS_CINR <15dB)
SeperatorThreshold2 = 0.001 * MaxEigenValue;                  % Good for RS-CINR > 15dB
Seperator1          = sum(EigenValues > SeperatorThreshold1); % Number of principle components
Seperator2          = sum(EigenValues > SeperatorThreshold2); % retained for the estimate

SignalComponentOfEstimate1  = zeros(NumCoordinates, 1);
NoiseComponentOfEstimate1   = zeros(NumCoordinates, 1);
SignalComponentOfEstimate2  = zeros(NumCoordinates, 1);
NoiseComponentOfEstimate2   = zeros(NumCoordinates, 1);

for EigIndex = 1:NumCoordinates
     EigenVector     = EigenVectors(:, EigIndex);
     Norm            = sum(RawFreqResponse .* conj(EigenVector)) / ...
                      (sum(conj(EigenVector).*EigenVector));
     if(EigIndex >  NumCoordinates - Seperator1)
         % Build the signal component of the frequency response. Note that the eigenvectors 
         % belonging to the signal components will have some noise as well. 
         SignalComponentOfEstimate1 = SignalComponentOfEstimate1 + Norm * EigenVector;
     else
         % Compute the noise belonging to those eigenvectors that do not belong to the signal 
         % component of the noisy raw frequency response. We use this to compute the RS-CINR.
         NoiseComponentOfEstimate1  = NoiseComponentOfEstimate1  + Norm * EigenVector; 
     end
    
     if(EigIndex >  NumCoordinates - Seperator2)
         SignalComponentOfEstimate2 = SignalComponentOfEstimate2 + Norm * EigenVector;
     else
         NoiseComponentOfEstimate2  = NoiseComponentOfEstimate2  + Norm * EigenVector; 
     end
end   

% Compute RS-CINR for Seperator1
TotalNoisePower1  = MeanSquare(NoiseComponentOfEstimate1)  ...
                    * NumCoordinates / (NumCoordinates - Seperator1);
TotalSignalPower1 = MeanSquare(SignalComponentOfEstimate1) ...
                    - TotalNoisePower1 * Seperator2/NumCoordinates;
RsCinrdB1         = 10*log10(TotalSignalPower1/TotalNoisePower1);

% Compute RS-CINR for Seperator2
TotalNoisePower2  = MeanSquare(NoiseComponentOfEstimate2)  ...
                    * NumCoordinates / (NumCoordinates - Seperator2);
TotalSignalPower2 = MeanSquare(SignalComponentOfEstimate1) ...
                    - TotalNoisePower1 * Seperator2/NumCoordinates;
RsCinrdB2         = 10*log10(TotalSignalPower2/TotalNoisePower2);
                 
% Only choose the results of seperator2 only if its RS-CINR results is truely better
if(real(RsCinrdB2) > (real(RsCinrdB1) + 1))
   RsCinrdB        = real(RsCinrdB2);    % Results for seperator1
   ChannelEstimate = SignalComponentOfEstimate2;
else 
   RsCinrdB        = real(RsCinrdB1);    % Results for seperator2
   ChannelEstimate = SignalComponentOfEstimate1;
end

figure(2);
subplot(2,1,1);
plot(Frequencies, real(FreqResponse), 'k', 'Linewidth', 2); grid on; hold on;
plot(Frequencies, real(RawFreqResponse), 'k:o', 'Markersize', 4);
plot(Frequencies, real(SignalComponentOfEstimate1), 'k--', 'Linewidth', 2);
title('Real Component of The Ideal, Raw and Estimated Frequency Response');
legend('Ideal (no Noise)', 'Raw (with Noise)', 'Estimate');
xlabel('Frequency in Hz');
axis([-1e7 1e7 -3 3]);
subplot(2,1,2);
plot(Frequencies, imag(FreqResponse), 'k', 'Linewidth', 2); grid on; hold on;
plot(Frequencies, imag(RawFreqResponse), 'k:o', 'Markersize', 4);
plot(Frequencies, imag(SignalComponentOfEstimate1), 'k--', 'Linewidth', 2);
title('Imaginary Component of The Ideal, Raw and Estimated Frequency Response');
legend('Ideal (no Noise)', 'Raw (with Noise)', 'Estimate');
xlabel('Frequency in Hz');
axis([-1e7 1e7 -3 3]);



function FreqResponse = ComputeFrequencyResponse(...
                           ChannelModel    ... % Multipath Channel Model
                         , f               ... % The frequencies and the time at which
                         , Time)               % to evaluate the Frequency Response

FreqResponse = zeros(size(f));
NumPaths     = size(ChannelModel, 1);      % Get number of rows of the matrix
 
for p = 1:NumPaths
   Cp      = ChannelModel(p, 1);           % Complex scalar
   Tau_p   = ChannelModel(p, 2);           % The delay 
   f_p     = ChannelModel(p, 3);           % The Doppler Shift
   FreqResponse = FreqResponse + Cp * exp(-1j*2*pi*Tau_p*f) .* exp(1j*2*pi*f_p*Time);
end     
end
