%% Text: Digital Signal Processing in Modern Communication Systems (Edition 2)
%% Chapter 7 - Section 1.3.5: This script will use Principle Component Analysis to reduce the amount of
%%              noise in the raw channel estimate of the 802.11a receiver.
%% Copyright: Andreas Schwarzinger May 2013
%%
%% Author: Andreas Schwarzinger  

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% File: ChannelEstimation_WLAN.m
% Explanation: This script will use Principle Component Analysis to reduce the amount of noise in the 
%              Raw Channel estimate of the 802.11a receiver.
% Author: Andreas Schwarzinger                                                       Date: 2.24.2020

clc;
clear variables;
close all;

%  A simply anonymous function to compute the mean square or total power of a sequence.
MeanSquare = @(x) (  mean( x(:).*conj(x(:)) )  );

%% 0. Some basic definitions
SubcarrierSpacing = 312.5e3;             % = 20e6/64
Subcarriers       =  -26:26;             % Only Subcarriers -26:-1 and 1:26 carry information. 
f                 = Subcarriers*SubcarrierSpacing;  % The frequencies of the subcarriers

%% 1.  Define the frequency response
%  1a. An example multipath configuration with 6 paths.
%               C_n         Delay_n
Path   =    [       1,         0; ...
                 -0.7j,   -20e-9; ...
            0.3 - 0.7j,    80e-9; ...
                    1j,   120e-9; ...
            0.4       ,   200e-9; ...
            0.1 - 0.1j,   250e-9];

FR_Ideal = ComputeFrequencyResponse(Path, f);

% Ensure that the rms value of the frequency response is unity. This simply ensure that
% the transmit and receive signal will have the same power.
MeanFRPower = MeanSquare(FR_Ideal);
FR_Ideal    = (1/sqrt(MeanFRPower))*FR_Ideal; % Normalize to unity power

figure(1);
subplot(2,1,1);
plot(f, real(FR_Ideal), 'k-o', 'Markersize', 5); grid on; hold on;
plot(f, imag(FR_Ideal), 'k-d', 'Markersize', 5);
xlabel('Subcarrier Frequency');
title('Real and Imaginary Portions of the Frequency Response of the Channel');
legend('Real','Imag');

%  1b. Compute the impulse response
IFFT         = ifft(FR_Ideal);
%  Reordering the impulse response so negative time appears first in the TimeResponse vector
%  and positive time last.
TimeResponse = [IFFT(1, 28:end), IFFT(1,1:27)];
Time         = (-26:26)/(SubcarrierSpacing*53);
 
subplot(2,1,2);
stem(Time,  (abs(TimeResponse)), 'k'); grid on; hold on;
xlabel('Time (sec)'); ylabel('dB');
title('Abs Value Channel Impulse Response');

Mask = [ones(1,6) zeros(1,44) ones(1, 3)];
FFT  = fft(IFFT.*Mask);
figure(22);
plot(f, real(FFT), 'r'); grid on; hold on;
plot(f, imag(FFT), 'b');



%% 2. Simulate the arrival of the two long training symbols
%     The 802.11a transmitter send two long training symbols. In the code below we will do the following:
%     a. Apply the frequency response to each training symbol to simulate the channel effect.
%     b. Add noise to each received training symbol.
%     c. Average the two.
%     d. Multiply the averge by the complex conjugate of the long training symbol values to compute
%        the received frequency response.

SNR_dB      = 0;               % In dB
SNR_Linear  = 10^(SNR_dB/10);  % dimensionless
disp(['Original SNR (dB): ' num2str(SNR_dB)]);

%     2a. The long training symbol is defined via the following values for the subcarrier range -26:26.
LongSymbol = [1 1 -1 -1 1 1   -1 1 -1 1   1 1 1 1   1 -1 -1 1   1 -1 1 -1   1 1 1 1 ... % -26:-1 
     0  1 -1 -1   1 1 -1 1   -1 1 -1 -1  -1 -1 -1 1  1 -1 -1 1   -1  1 -1 1    1 1 1];  % 0:26   
 
%     Apply the frequency response to the long symbol and compute the signal power (which is 1)
LongSymbol_RxIdeal = LongSymbol .* FR_Ideal;
SignalPower        = (1/52) * (LongSymbol_RxIdeal * LongSymbol_RxIdeal'); 

%     2b. Add noise in accordance with the SNR specified to simulate two received long symbols
NoisePower         = SignalPower/SNR_Linear;
LongSymbol_Rx1     = LongSymbol_RxIdeal + 0.707*sqrt(NoisePower)*(randn(1,53) + 1j*randn(1,53));
LongSymbol_Rx2     = LongSymbol_RxIdeal + 0.707*sqrt(NoisePower)*(randn(1,53) + 1j*randn(1,53));

%     2c/d. Average the two and derotate the result to produce an estimate of the frequency response.
%           Without the additional noise removal, we would be stuck with this situation.
LongSymbol_Rx      = 0.5*LongSymbol_Rx1 + 0.5*LongSymbol_Rx2;
FR_Estimate        = LongSymbol_Rx .* conj(LongSymbol);

% Luckily, we know what the actual frequency response is, and we can therefore subtract it to 
% determine how much noise is left and recompute the SNR. It will be 3dB better
RemainingNoise     = FR_Estimate - FR_Ideal;
NewNoisePower      = MeanSquare(RemainingNoise);
New_SNR_dB         = 10*log10(SignalPower/NewNoisePower);  
disp(['SNR after averaging (dB): ' num2str(New_SNR_dB)]);


%% 3.  Cleaning up the frequency estimate even more using the principle component analysis
%  3a. Creating a Data Set for the PCA technique to act upon.
%      We will simulate and record all frequency responses that may befall the channel.
%      Here we list all delays that the paths in the system could assume.
Delay                =   -50e-9:2e-9:350e-9;
NumChannelConditions = length(Delay); 

% FR_IdealMatrix will hold the frequency responses for each channel condition.
FR_IdealMatrix   = zeros(length(f), NumChannelConditions);
 
for ChannelIndex = 0:NumChannelConditions-1
    StartAngle = rand(1,1)*2*pi;  % We must make this random (otherwise the computation fails)
    CurrentDelay                      = Delay(1, ChannelIndex + 1);
    Current_FR_Ideal                  = exp(-1j*(2*pi*f*CurrentDelay + StartAngle)).';
    FR_IdealMatrix(:, ChannelIndex+1) = Current_FR_Ideal;    
end

MeanFR_Ideal = mean(FR_IdealMatrix, 2);


% --------------------------------------------------------------------------------------------
%  3b: Compute the Covariance Matrix
CovarianceM = cov(FR_IdealMatrix');      % The covariance matrix must have dimensions  
[V, D]      = eig(CovarianceM);              % The eigenvalue decomposition

% Get the eigenvalues
NumberEig   = length(V); 
EigenValues = zeros(1, NumberEig);
for i = 1:NumberEig; EigenValues(1, i) = D(i,i); end 
ShowEigenValues = true;
if(ShowEigenValues == true)
    figure(2);
    stem(EigenValues, 'k'); grid on; hold on;
    title('The size of each Eigenvalue'); 
    xlabel('Eigenvalue index');
    legend('Eigen Values');
    disp('Look at the eigenvalue sizes and determine the number of eigenvectors to retain in the exercise.');
end

% Get the eigen vectors. The eigenvectors are extracted as follows.
% EigenV1      = V(:, 1);
% EigenV2      = V(:, 2);
%    :
% EigenV_Max   = V(:, end);

RelevantEigenvectors = 9;
disp(['-> The power seperator (number of eigen vectors to retain) is set to ' num2str(RelevantEigenvectors)]);
disp('-> This should be the number of eigen vectors along which the majority of signal power is available.');

% ------------------------------------------------------------------------------------------------
% 3c. Use projection to see how it extends along the eigen vectors of the new space we have computed.
%
% ------------------------------------------------------------------------------------------------
Range                 = NumberEig - RelevantEigenvectors + 1:NumberEig;
RemainingEigenValues  = EigenValues(1, Range);
RemainingEigenVectors = V(:, Range);

Final_FR_Estimate = zeros(53, 1);
for i = 1:RelevantEigenvectors
    VEig              = RemainingEigenVectors(:, i);
    CurrentProjection = dot(VEig, FR_Estimate)/dot(VEig, VEig) * VEig;
    Final_FR_Estimate = Final_FR_Estimate + CurrentProjection;
end

RemainingNoise     = FR_Ideal.' - Final_FR_Estimate;
NewNoisePower      = MeanSquare(RemainingNoise);
New_SNR_dB         = 10*log10(SignalPower/NewNoisePower);

disp(['Final SNR (dB): ' num2str(New_SNR_dB)]);

figure(3);
subplot(2,1,1);
plot(f, real(FR_Ideal), 'k', 'Linewidth', 2); grid on; hold on;   
plot(f, real(LongSymbol_Rx2 .* conj(LongSymbol)), 'k-.', 'Linewidth', 2); 
axis([-8.5e6 8.5e6 -1 2]);
xlabel('Subcarrier Frequency');
title('Real Portion of the Actual Frequency Response and the original Dirty Estimate of the Frequency Response');
legend('Ideal Frequency Response','Frequency Response without Processing')

subplot(2,1,2);
plot(f, real(FR_Ideal), 'k', 'Linewidth', 2); grid on; hold on;  
plot(f, real(Final_FR_Estimate), 'k-.', 'Linewidth', 2); 
axis([-8.5e6 8.5e6 -1 2]);
xlabel('Subcarrier Frequency');
title('Real Portion of the Actual Frequency Response and Final Estimate of the Frequency Response');
legend('Ideal Frequency Response', 'Final Frequency Response Estimate')











function FR_Ideal = ComputeFrequencyResponse(Path, f)

FR_Path1 = Path(1,1) * exp(-1j*2*pi*f*Path(1,2)); 
FR_Path2 = Path(2,1) * exp(-1j*2*pi*f*Path(2,2)); 
FR_Path3 = Path(3,1) * exp(-1j*2*pi*f*Path(3,2)); 
FR_Path4 = Path(4,1) * exp(-1j*2*pi*f*Path(4,2)); 
FR_Path5 = Path(5,1) * exp(-1j*2*pi*f*Path(5,2)); 
FR_Path6 = Path(6,1) * exp(-1j*2*pi*f*Path(6,2)); 
FR_Ideal = FR_Path1 + FR_Path2 + FR_Path3 + FR_Path4 + FR_Path5 + FR_Path6;
end