% File: Lte Modulation / Demodulation Simulation
% Author: Andreas Schwarzinger                  Date: 8.2.2021

clc;
clear variables;
close all;

% A resource grid for a 1.4MHz Lte Signal 
NDLRB                  = 6;     % Number of downlink resource blocks
NumSubcarriersPerRB    = 12;    % There are 12 subcarriers per resource block
SubcarrierSpacing      = 15000; % Hz
CP_Selection           = 1;     % Normal cyclic prefix
SubframeStartTime      = 0;     % The start time of the subframe to demodulate
TimeAdvance            = 1e-6;  % Start FFT 1 microsecond early (inside CP)
NumSubframes           = 1;     % We will simulate only one subframe
NumSubcarrier          = NDLRB * 12;  

ResourceGridTx         = zeros(NumSubcarrier, 14*NumSubframes); % NDLRB = 72
k                      = 35:36;                     % Subcarrier indices [35, 36]
ResourceGridTx(k+1, :) = ones(2, 14*NumSubframes);  % Populate only those subcarriers

% Test 1: Generate the 1.4MHz Waveform at all FFT sizes and associated sample rates)
for FftSize = [128, 256, 512, 1024, 1408, 1536, 2048]
    [Waveform, SampleRate] = ModulateLteSignal(ResourceGridTx, 512);
    ResourceGridRx = DemodulateLteSignal(Waveform  ...  % Vector of complex samples
                                    , NumSubframes...  % Number subframes to demodulate
                                    , NDLRB       ...  % Number download resource blocks
                                    , CP_Selection...  % Cyclic Prefix 1/2 - normal/ext
                                    , SubframeStartTime   ... % The subframe start time
                                    , TimeAdvance ...
                                    , SampleRate);    
                                
    MaxError = max(abs(ResourceGridRx - ResourceGridTx));
    if(MaxError > 1e-5); error('Maximum error exceeds limit'); end
end

% Test 2: Generate the same 1.4MHz LTE waveform upsampling by a factor of 4
% FFT size = 512 rather than 128 (Sample rate will be 7.68e6 Hz)
[WaveformTX, SampleRate] = ModulateLteSignal(ResourceGridTx, 512);

% Time advancing waveform by 1 sample then decimate the waveform by a factor of two
% New sample rate will be 3.84MHz (thus FFT size = 256)
WaveformRX               = WaveformTX(1, 2:2:end);
NewSampleRate            = SampleRate / 2;

% Demodulate the Lte Waveform
SubframeStartTime =  -1 / SampleRate;  % Now starts earlier 
TimeAdvance       = 1e-6;              % Take the FFT TimeAdvance seconds early in time

ResourceGridRx = DemodulateLteSignal(WaveformRX  ...  % Vector of complex samples
                                    , NumSubframes...  % Number subframes to demodulate
                                    , NDLRB       ...  % Number download resource blocks
                                    , CP_Selection...  % Cyclic Prefix 1/2 - normal/ext
                                    , SubframeStartTime   ... % The subframe start time
                                    , TimeAdvance ...
                                    , NewSampleRate);  % 256 * 15000 = 3.84MHz
                                
 MaxError = max(abs(ResourceGridRx - ResourceGridTx));
 if(MaxError > 1e-5); error('Maximum error exceeds limit'); end