function [ OutputWaveform, SampleRate ] = ModulateLteSignal(ResourceGrid, NewFffSize)
% This function will render an LTE resource grids into the time domain.
% Author: Andreas Schwarzinger                          Date: 8/12/2020

% Input:  ResourceGridArray - dimensions (NDLRB*12, NumberFrames * 10 * 14) (CP normal)
%                                        (NDLRB*12, NumberFrames * 10 * 12) (CP extended)
%         NewFffSize        - allows us to render the waveform at a higher sample rate
%                     The FFT size must be the minimum size for the given resource grid.
%                     NDLRB 6   - Possible FFT sizes = 128, 256, 512, 1024, 1408, 1536, 2048
%                     NDLRB 15  - Possible FFT sizes =      256, 512, 1024, 1408, 1536, 2048
%                     NDLRB 25  - Possible FFT sizes =           512, 1024, 1408, 1536, 2048
%                     NDLRB 50  - Possible FFT sizes =                1024, 1408, 1536, 2048
%                     NDLRB 75  - Possible FFT sizes =                      1408, 1536, 2048
%                     NDLRB 100 - Possible FFT sizes =                      1408, 1536, 2048

% Output: OutputWaveform    - dimensions (1, :)
%         SampleRate        - the sample rate of the waveform calculated

assert(nargin == 1 || nargin == 2, 'We may have either 1 or 2 input arguments.');
if(nargin == 1);      N_FFT_Override = 0;
elseif(nargin == 2) % Check to see whether the NewFftSize is valid.
   ProperFftSize = sum(NewFffSize == [128, 256, 512, 1024, 1408, 1536, 2048]) == 1; 
   assert(ProperFftSize == 1, 'Invalid FFT size.');
   N_FFT_Override = NewFffSize;
else;                  error('We may have either 1 or 2 input arguments.');
end

%% Step 1: Check input argument for proper size
%  1a. Find number of resource blocks in signal
[NumSubcarriers, NumOfdmSymbols]   = size(ResourceGrid);

% 1b. Determine number of downlink resource blocks (or NDLRB)
NDLRB          = NumSubcarriers/12;
if(mod(NumSubcarriers, 12) ~= 0)
    error('RenderToTimeDomain.m: Bad input argument dimension.'); end 

% 1c. Determine the cyclic prefix, given that there must be eithe 14 or 12 symbols per subframe.
NormalCP       = mod(NumOfdmSymbols, 14) == 0;
ExtendedCP     = mod(NumOfdmSymbols, 12) == 0;
if(~NormalCP && ~ExtendedCP)    
    error('RenderToTimeDomain.m: Bad input argument dimension.'); end

% 1d. Find the number of subframes in the signal.
if(NormalCP == 1)
    NumberOfSubframes = NumOfdmSymbols/14;
elseif(ExtendedCP == 1)
    NumberOfSubframes = NumOfdmSymbols/12;
else
    error('RenderToTimeDomain.m: Number of Ofdm Symbols could not be determined.');
end


%% Step 2: Initialize the output variable
% Determine desired FFT size
switch(NDLRB)
    case   6; N = max(128,  N_FFT_Override); % N 128  -> Ncp=[10   9*ones(1, 6)]; Ecp + 32*ones(1,6);
    case  15; N = max(256,  N_FFT_Override); % N 256  -> Ncp=[20  18*ones(1, 6)]; Ecp + 64*ones(1,6);
    case  25; N = max(512,  N_FFT_Override); % N 512  -> Ncp=[40  36*ones(1, 6)]; Ecp + 128*ones(1,6);
    case  50; N = max(1024, N_FFT_Override); % N 1024 -> Ncp=[80  72*ones(1, 6)]; Ecp + 256*ones(1,6);
    case  75; N = max(1408, N_FFT_Override); % N 1408 -> Ncp=[110 99*ones(1, 6)]; Ecp + 352*ones(1,6);
    case 100; N = max(1408, N_FFT_Override); % N 1536 -> Ncp=[120 108*ones(1, 6)];Ecp + 384*ones(1,6);  
                                             % N 2048 -> Ncp=[160 144*ones(1, 6)];Ecp + 512*ones(1,6);  
    otherwise; error('RenderToTimeDomain: Invalid value for NDLRB (6, 15, 25, 50, 75, 100)');     
end

% Determine number of CP samples for normal and extended cyclic prefix
Multiplier = N / 128;
Ncp = Multiplier * [ 10   9*ones(1, 6)]; Ecp =  Multiplier * 32*ones(1,6);

% The second output argument is now known
SubcarrierSpacing = 15000;        % Hz
SampleRate = N * SubcarrierSpacing;

if(NormalCP == 1)
    OutputLength        = NumberOfSubframes * (N*14 + 2*Ncp(1,1) + 12*Ncp(1,2));
    NsymbolsPerSlot     = 7;
    CPLengths           = Ncp;
else
    OutputLength        = NumberOfSubframes * (N*12 + 12*Ecp(1,1));
    NsymbolsPerSlot     = 6;
    CPLengths           = Ecp;
end
OutputWaveform = zeros(1, OutputLength);

%% Step 3: Build the input mapping for the IFFT
PositiveSubcarriers = NumSubcarriers/2 : (NumSubcarriers - 1);  % Index in resource grid
NegativeSubcarriers = 0:(NumSubcarriers/2 - 1);                 % Index in resource grid

PosFreqIndicesIFFT  = 1:NumSubcarriers/2;                       % Index in IFFT
NegFreqIndicesIFFT  = (N - NumSubcarriers/2):(N-1);             % Index in IFFT

StartIndex = 0;
for SymbolIndex = 0:NumOfdmSymbols-1
    % Compute IFFT output
    IFFT_Input                            = zeros(N, 1);
    IFFT_Input(PosFreqIndicesIFFT + 1, 1) = ResourceGrid(PositiveSubcarriers + 1, SymbolIndex + 1);
    IFFT_Input(NegFreqIndicesIFFT + 1, 1) = ResourceGrid(NegativeSubcarriers + 1, SymbolIndex + 1);
    IFFT_Output                           = N*ifft(IFFT_Input, N);   
                                                                    
    % Generate cyclic prefix
    SymbolIndexInSlot                     = mod(SymbolIndex, NsymbolsPerSlot); 
    CurrentCpLength                       = CPLengths(1, SymbolIndexInSlot + 1); 
    CyclicPrefix                          = IFFT_Output(N-CurrentCpLength+1:N, 1);
    
    % Generate Output OFDM symbol by prepending the cyclic prefix
    OutputSymbol                          = [CyclicPrefix; IFFT_Output];
    
    % Place output symbols into output waveform array
    Range                                 = StartIndex+1:StartIndex+length(OutputSymbol);
    OutputWaveform(1, Range)              = OutputSymbol.';
    StartIndex                            = StartIndex + length(OutputSymbol);
end
assert(StartIndex == OutputLength, 'The output waveform array was not completely filled');
end

