%% Text: Digital Signal Processing in Modern Communication Systems (Edition 2)
%% Chapter 7 - Section 4: The following function converts LTE information in a resource grid into 
%%                        a time domain waveform. It is the OFDM modulator.
%% Copyright: Andreas Schwarzinger Nov 2021
%%
%% Author: Andreas Schwarzinger                              Nov 2021


function [ OutputWaveform, SampleRate ] = RenderToTimeDomainBook( ResourceGridArray )
% This function will render an LTE resource grids into the time domain.
% Author: Andreas Schwarzinger                          Date: 2/16/2020

% Input:  ResourceGridArray  - dimensions (NDLRB*12, NumberFrames * 10 * 14)
%         AlwaysUse2048FFT   - if true, then the output waveform is always rendered at 30.72MHz,
 
% Output: OutputWaveform     - dimensions (1, :)

if(nargin == 1)
    Use_2048Times_X_IFFT = false;
end

%% Step 1: Check input argument for proper size
%  1a. Find number of resource blocks in signal
[NumSubcarriers, NumOfdmSymbols]   = size(ResourceGridArray);

% 1b. Determine number of downlink resource blocks
NDLRB          = NumSubcarriers/12;
if(mod(NumSubcarriers, 12) ~= 0); error('RenderToTimeDomain.m: Bad input argument dimension.'); end 

% 1c. Find number of frames in signal and associated cyclic prefix size
NormalCP       = mod(NumOfdmSymbols, 14) == 0;
ExtendedCP     = mod(NumOfdmSymbols, 12) == 0;

if(~NormalCP && ~ExtendedCP); error('RenderToTimeDomain.m: Bad input argument dimension.'); end

if(NormalCP == 1)
    NumberOfSubframes = NumOfdmSymbols/14;
elseif(ExtendedCP == 1)
    NumberOfSubframes = NumOfdmSymbols/12;
else
    error('RenderToTimeDomain.m: Number of Ofdm Symbols could not be determined.');
end


%% Step 2: Initialize the output variable
switch(NDLRB)
    case   6;  N =  128; Ncp = [ 10   9*ones(1, 6)]; Ecp =  32*ones(1,6);
    case  15;  N =  256; Ncp = [ 20  18*ones(1, 6)]; Ecp =  64*ones(1,6);
    case  25;  N =  512; Ncp = [ 40  36*ones(1, 6)]; Ecp = 128*ones(1,6);
    case  50;  N = 1024; Ncp = [ 80  72*ones(1, 6)]; Ecp = 256*ones(1,6);
    case  75;  N = 1408; Ncp = [110  99*ones(1, 6)]; Ecp = 352*ones(1,6);
    case 100;  N = 1408; Ncp = [110  99*ones(1, 6)]; Ecp = 352*ones(1,6);
    otherwise; error('RenderToTimeDomain: Invalid value for NDLRB (6, 15, 25, 50, 75, 100)');     
end

if(Use_2048Times_X_IFFT ~= 0); N = X*2048; Ncp = X*[160 144*ones(1, 6)]; Ecp = X*512*ones(1,6); end
% The second output argument is now known
SubcarrierSpacing = 15000;   % Hz
SampleRate = N * SubcarrierSpacing;

if(NormalCP == 1)
    OutputLength        = NumberOfSubframes * (N*14 + 2*Ncp(1,1) + 12*Ncp(1,2));
    NsymbolsPerSlot     = 7;
    CPLengths           = Ncp;
else
    OutputLength        = NumberOfSubframes * (N*12 + 12*Ecp(1,1));
    NsymbolsPerSlot     = 6;
    CPLengths           = Ecp;
end
OutputWaveform = zeros(1, OutputLength);

%% Step 2: Build the input mapping for the IFFT
PositiveSubcarriers = NumSubcarriers/2 : (NumSubcarriers - 1);  % Index in resource grid
NegativeSubcarriers = 0:(NumSubcarriers/2 - 1);                 % Index in resource grid

PosFreqIndicesIFFT  = 1:NumSubcarriers/2;                       % Index in IFFT
NegFreqIndicesIFFT  = (N - NumSubcarriers/2):(N-1);             % Index in IFFT

StartIndex = 0;
for SymbolIndex = 0:NumOfdmSymbols-1
    % Compute IFFT output
    IFFT_Input                            = zeros(N, 1);
    IFFT_Input(PosFreqIndicesIFFT + 1, 1) = ResourceGridArray(PositiveSubcarriers + 1, SymbolIndex + 1);
    IFFT_Input(NegFreqIndicesIFFT + 1, 1) = ResourceGridArray(NegativeSubcarriers + 1, SymbolIndex + 1);
    IFFT_Output                           = N*ifft(IFFT_Input, N);  % Scaling by 1 guaranteed same 
                                                                    % same output scaling for all FFT sizes
    % Generate cyclic prefix
    SymbolIndexInSlot                     = mod(SymbolIndex, NsymbolsPerSlot); 
    CurrentCpLength                       = CPLengths(1, SymbolIndexInSlot + 1); 
    CyclicPrefix                          = IFFT_Output(N-CurrentCpLength+1:N, 1);
    
    % Generate Output symbol
    OutputSymbol                          = [CyclicPrefix; IFFT_Output];
    
    % Place output symbols into output waveform array
    Range                                 = StartIndex+1:StartIndex+length(OutputSymbol);
    OutputWaveform(1, Range)              = OutputSymbol.';
    StartIndex                            = StartIndex + length(OutputSymbol);
end
assert(StartIndex == OutputLength, 'The output waveform array was not completely filled');
end

