%% Text: Digital Signal Processing in Modern Communication Systems (Edition 2)
%% Chapter 7 - Section 6: This function demodulates a baseband downlink LTE Ofdm signal.
%%             regression to find the receiver start time of all OFDM symbols beyond the first 10 coordinates
%% Copyright: Andreas Schwarzinger May 2021

% Input Arguments: 
% -> Waveform     -> Baseband IQ signal
% -> NumSubfarmes -> The number of subfarmes to demodulate. the IQ waveform must be long enough.
% -> NDLRB        -> The number of downlink resource blocks (The signal bandwidth)
% -> CP_Selection -> Either normal (1) or extended 92)
% -> StartTime    -> The FFT start time computed by the LTE receiver during timing aquisition.

function ResourceGrid = DemodulateLteSignal (Waveform    ...  % Vector of complex samples
                                          , NumSubframes...  % Number subframes to demodulate
                                          , NDLRB       ...  % Number download resource blocks
                                          , CP_Selection...  % Cyclic Prefix 1/2 - normal/ext
                                          , StartTime)
                                      
% Set parameters and some error checking    
SubcarrierSpacing = 15000;  % Hz                
TimeIfftSection   = 1/SubcarrierSpacing; % Length (sec) of IFFT portion of OFDM symbol = 66.667usec                                                
                           
switch(NDLRB)                                       
    case   6; SampleRate =  1.92e6; FFT_Size =  128;   % Corresponds to BW =  1.4MHz 
    case  15; SampleRate =  3.84e6; FFT_Size =  256;   % Corresponds to BW =  3.0MHz
    case  25; SampleRate =  7.68e6; FFT_Size =  512;   % Corresponds to BW =  5.0MHz
    case  50; SampleRate = 15.36e6; FFT_Size = 1024;   % Corresponds to BW = 10.0MHz
    case  75; SampleRate = 21.12e6; FFT_Size = 1408;   % Corresponds to BW = 15.0MHz
    case 100; SampleRate = 30.72e6; FFT_Size = 2048;   % Corresponds to BW = 20.0MHz
    otherwise; error('Unsupported number of downlink resource blocks.');
end

switch(CP_Selection)       % First Symbol in slot  -  All other symbols in slot]
    case 1; CP_Times       = [10/1.92e6,              9/1.92e6];  % seconds - Normal CP
            SymbolsPerSlot = 7;
    case 2; CP_Times       = [32/1.92e6,             32/1.92e6];  % seconds - Extended CP
            SymbolsPerSlot = 6;
    otherwise; error('Unsupported cyclic prefix configuration.');
end
                 
NumberOfPositiveSubcarriers = 12 * NDLRB / 2;    % Each resource block consists of 12 subcarriers
                              % Half the subcarriers will be on the positive and half on the negative
                              % frequency side. There is no information at the DC subcarrier.
                              
% Resource grid where we deposit the FFT output
ResourceGrid = zeros(NDLRB*12, SymbolsPerSlot*2);

FFT_Time             = StartTime;
TotalNumberOfSymbols = NumSubframes * 2 * SymbolsPerSlot; % There are two slots / subframe 

for Symbol = 0:TotalNumberOfSymbols-1       
    if(mod(Symbol, SymbolsPerSlot) == 0)
        CP_Time = CP_Times(1,1);
    else
        CP_Time = CP_Times(1,2);
    end
    FFT_Time      = FFT_Time + CP_Time;
    EarlySample   = floor(FFT_Time * SampleRate); 
    % We are taking the FFT of samples that are happen earlier in time.
    % To the FFT it looks like we are transforming a delayed signal.
    ApparentDelay = FFT_Time - EarlySample/SampleRate;
    
    % Take the following samples and FFT transform them
    Range         = EarlySample:(EarlySample + FFT_Size - 1);
    Samples       = Waveform(1, Range + 1);
    FFT_Output    = fft(Samples, FFT_Size)/FFT_Size;
  
    % Adjust the FFT output phases given the Apparent Delay
    Pos_Frequencies = (1:1:NumberOfPositiveSubcarriers)*15000;  % Remember -> skip DC
    Pos_Index_Range =  1:1:NumberOfPositiveSubcarriers;     
    FFT_Output(1, Pos_Index_Range + 1) = ...
    FFT_Output(1, Pos_Index_Range + 1) .* exp(1j*2*pi*ApparentDelay*Pos_Frequencies);
    
    Neg_Frequencies = (-NumberOfPositiveSubcarriers:1:-1) *15000;
    Neg_Index_Range = FFT_Size - NumberOfPositiveSubcarriers:1:FFT_Size - 1;
    FFT_Output(1, Neg_Index_Range + 1) = ...
    FFT_Output(1, Neg_Index_Range + 1) .* exp(1j*2*pi*ApparentDelay*Neg_Frequencies);
    
    % Map the FFT output into the resource grid
    Pos_Range = NumberOfPositiveSubcarriers:2*NumberOfPositiveSubcarriers - 1;   
    Neg_Range = 0:1:NumberOfPositiveSubcarriers-1;  
    ResourceGrid(Pos_Range + 1, Symbol + 1) = FFT_Output(1, Pos_Index_Range + 1);
    ResourceGrid(Neg_Range + 1, Symbol + 1) = FFT_Output(1, Neg_Index_Range + 1);
        
    % Increment FFT timing instant
    FFT_Time      = FFT_Time + TimeIfftSection;
end