%% Text: Digital Signal Processing in Modern Communication Systems (Edition 2)
%% Chapter 7 - Section 6: The following scripts implements weighted least squares optimization during linear 
%%             regression to find the receiver start time of all OFDM symbols beyond the first 10 coordinates
%% Copyright: Andreas Schwarzinger February 2021


clc;
clear variables;
format long;

% --------------------------------------------------------------------------
% Define the time base error
TxErrorPpm            =  2;                       % Transmit time base error
RxErrorPpm            =  0;                       % Receiver time base error
TotalTimeBaseErrorPpm = RxErrorPpm - TxErrorPpm;  

% Define the actual frequency offset produced due to the total time base error.
CarrierFreq           = 2e9;                                          % Hertz
ActualFreqOffset      = -CarrierFreq * TotalTimeBaseErrorPpm * 1e-6;    % Hertz

% -------------------------------------------------------------------------
% Approximate the TxTime and RxTime coordinates of the cross-correlation peaks
% produced by the total time base error configuration above. Experiment with the
% ObservationPeriod.
Slope                 = 1 + TotalTimeBaseErrorPpm * 1e-6;
ObservationPeriod     = 200e-3;                        
TxTime                = (0:0.0005:ObservationPeriod)'; % Seconds

% The RxTime coordinates have an intial value at TxTime = 0. We add this offset
% as well as inaccuracies when measuring the receive time, RxTime, at the 
% correlation peaks. Experiment with the size of the MaxRxTimeInaccuracy.
RxTimeOffset          = 0.2341;                        % The intercept point. 
MaxRxTimeInaccuracy   = 100e-9;                        % Seconds
RxTimeNoise           = MaxRxTimeInaccuracy * (2*rand(length(TxTime), 1) - 1);
RxTime                = TxTime * Slope + RxTimeOffset + RxTimeNoise;

figure(1);
subplot(2,1,1);
plot(TxTime, RxTime, 'k-o'); grid on; 
title('Receive Time versus Transmit Time Coordiantes');
xlabel('Tx Time (sec)');
ylabel('Rx Time (sec)');
subplot(2,1,2);
plot(TxTime, RxTimeNoise, 'k'); grid on;
title('Receive Time Inaccuracies');
xlabel('TxTime');
ylabel('RxTime Error');

% ----------------------------------------------------------------------
% We have the observed correlation peak time coordinates now as (TxTime, RxTime).
% It is our goal to compute the linear regression to estimate the intercept point
% and slope of the line describing the actual correlation peak time coordinates.
F      = [ones(size(TxTime)), TxTime];  % See Section 3.1.2 for more details
Weight = 1:length(TxTime); % Weigh the later coordinates more than the eariler ones.
W      = diag(Weight);
AW     = inv(F' * W * F) * F' * W * RxTime;  % F'*W*F is a simple 2x2 matrix

% AW represents the intercept point and the slope of a line
DetectedInterceptPoint       = AW(1,1);
DetectedSlope                = AW(2,1);

% We can now estimte the frequency offset that must exist in the signal
EstimatedTotalBaseErrorPpm   = (DetectedSlope - 1) * 1e6;
EstimatedFreqOffset          = -CarrierFreq * EstimatedTotalBaseErrorPpm * 1e-6;

% We can also estimate the receive time at which OFDM symbols starts.
% For the extended CP, the LTE OFDM symbol period is 1e-3/ 12 = 83.33333e-6 seconds
% We will figure out the Ofdm symbol start times for the last 50 milliseconds.
OfdmSymbolPeriod  = 1e-3 / 12;
TxTimeSymbolStart = 0:OfdmSymbolPeriod:ObservationPeriod;
RxTimeSymbolStart = DetectedInterceptPoint + DetectedSlope * TxTimeSymbolStart;

disp(['Actual    Frequency Offset (Hz): ' num2str(ActualFreqOffset) ]);
disp(['Estimated Frequency Offset (Hz): ' num2str(EstimatedFreqOffset) ]);
disp(['Remaining Frequency Offset (Hz): ' num2str(ActualFreqOffset - EstimatedFreqOffset)]);




 









