%% Text: Digital Signal Processing in Modern Communication Systems (Edition 2)
%% Chapter 7 - Section 7: This function implements the OFDM modulation component of the C-V2X transmitter.
%% Copyright: Andreas Schwarzinger February 2021

clc;
clear variables;
close all;


%% Set the FFT size.
%  Typical FFT sizes can be 1024 (10MHz case only), 1408, 1536, 2048 (10MHz or 20MHz cases)
FftSize = 1024;
ResourceGridArray = ones(600, 14);
[SampleRate, TxSequence] = CV2xModulate(ResourceGridArray, FftSize);
T = 1;
% ------------------------------------------------------------------------------------ 
%% -> OfdmModulate()
% ------------------------------------------------------------------------------------ 
% Brief:   CV2xModulate() function converts the ResourceGridArray into an IQ sequence.
% Inputs:  ResourceGridArray -> The resource grid to be OFDM modulated.
%          FftSize           -> Usually (N = 1024 -> 10MHz | 1408/1536/2048 -> 20MHz) 
% Outputs: SampleRate        -> Self Explanatory
%          TxSequence        -> The Ofdm modulated output sequence.
% Specs:  See TS36.211 Section 5.6
function [SampleRate, TxSequence] =  CV2xModulate(ResourceGridArray, FftSize) 
							
    [NumSubcarriers, NumOfdmSymbols]   = size(ResourceGridArray);    
    ProperNumSubcarriers = NumSubcarriers == 600 || NumSubcarriers == 1200;
    ProperNumSymbols     = mod(NumOfdmSymbols, 14) == 0;
    assert(ProperNumSubcarriers, 'We need to have either 600 or 1200 subcarriers.');
    assert(ProperNumSymbols,     'We to have an integer number of subframes.');
    
    % 1. Setup parameters 
    N_FFT_Target = FftSize;
    SampleRate   = FftSize * 15000;     % 15000Hz is the subcarrier spacing
    
    % The number of CP samples of the first and remaining symbols in a slot
    Ncp_Target   = [10, 9, 9, 9, 9, 9, 9] * FftSize/128; % Correct of Normal CP                                  
          
    NumSubframes = NumOfdmSymbols/14;   % There are 14 Ofdm Symbols per subframe
    NumTxSamples = NumSubframes * ...
	               (N_FFT_Target*14 + 2*Ncp_Target(1,1) + 12*Ncp_Target(1,2));
    TxSequence   = zeros(1, NumTxSamples);
          
    % 2. Build the input mapping for the IFFT (Index in resource grid)
    PositiveSubcarriers = NumSubcarriers/2 : (NumSubcarriers - 1);   
    NegativeSubcarriers = 0:(NumSubcarriers/2 - 1);                  
          
    % Index in IFFT (Different than the downlink -> 1:NumSubcarriers/2)
    PosFreqIndicesIFFT  = 0:NumSubcarriers/2 -1;  
    NegFreqIndicesIFFT  = (N_FFT_Target - NumSubcarriers/2):(N_FFT_Target-1);   
          
    % 3. Ofdm Modulate each symbol (See TS36.211 Section 5.6)
    StartIndex = 0;
    for SymbolIndex = 0:NumOfdmSymbols-1
        % 3a. Fill the  IFFT input buffer
        IFFT_Input                            = zeros(N_FFT_Target, 1);
        IFFT_Input(PosFreqIndicesIFFT + 1, 1) = ...
		          ResourceGridArray(PositiveSubcarriers + 1, SymbolIndex + 1);
        IFFT_Input(NegFreqIndicesIFFT + 1, 1) = ...
		          ResourceGridArray(NegativeSubcarriers + 1, SymbolIndex + 1);
                          
        % 3b. Take the IFFT
        IFFT_Output              = N_FFT_Target*ifft(IFFT_Input, N_FFT_Target); 
                                   % Scaling by N_FFT_Target guarantees same 
                                   % output scaling for all FFT sizes
        % 3c. Generate cyclic prefix
        SymbolIndexInSlot  = mod(SymbolIndex, 7);  % CP - normal (7 Symbols per slot)
        CurrentCpLength = Ncp_Target(1, SymbolIndexInSlot + 1);
        CyclicPrefix    = IFFT_Output(N_FFT_Target-CurrentCpLength+1:N_FFT_Target, 1); 
              
        % 3d.Generate time domain OFDM symbol including CP and 7500Hz shift
        OutputSymbol             = [CyclicPrefix; IFFT_Output];
        n                        = (-CurrentCpLength:N_FFT_Target-1).';
        f                        = 0.5 * (1/N_FFT_Target);
        OutputSymbol             = OutputSymbol.*exp(1j*2*pi*n*f);
              
        % 3e. Place output symbols into output waveform array
        Range                    = StartIndex+1:StartIndex+length(OutputSymbol);
        TxSequence(1, Range)     = OutputSymbol.';
        StartIndex               = StartIndex + length(OutputSymbol);
    end
end 
