%% Text: Digital Signal Processing in Modern Communication Systems (Edition 2)
%% Chapter 8 - Section 1: This script demonstrates optimal receive diversity combining using Maximum ratio 
%%                        combining and estimation viat the minimum mean squared error estimator.
%%                        See Example 8.1 in the book.
%% Copyright: Andreas Schwarzinger Oct 2021
%%
%% Author: Andreas Schwarzinger                              Oct 2021

clc;
clear variables;
close all;

%% Using the MMSE Estimator
% Start out by defining the transmit symbol power and noise powers at each antenna.
SigmaX  =    1;      % The signal power (we will use a simple QPSK symbol sequence)
SigmaV1 = 0.01;      % The noise power at antenna1
SigmaV2 = 0.01;      % The noise power at antenna2
SigmaV3 = 0.01;      % The noise power at antenna3

% Let us now define the channel coefficients describing the frequency response for each antenna path.
h1      =  1 + 1j;      % Channel coefficients for path 1
h2      =  -0.2 - 1.3j; % Channel coefficients for path 2
h3      =  -1j*0.5;     % Channel coefficients for path 3

% The linear signal to noise ratios at each observation are as follows.
SNR1    =  SigmaX*h1*conj(h1)/SigmaV1;  disp(['Linear SNR at y1 = ' num2str(SNR1)]);
SNR2    =  SigmaX*h2*conj(h2)/SigmaV2;  disp(['Linear SNR at y2 = ' num2str(SNR2)]);
SNR3    =  SigmaX*h3*conj(h3)/SigmaV3;  disp(['Linear SNR at y3 = ' num2str(SNR3)]);

% Let's now build the different matrices needed to solve for the coefficient of our MMSE estimator.
Rx = SigmaX;
Rv = [SigmaV1,       0,        0; ...
      0,       SigmaV2,        0; ...
      0,             0,  SigmaV3];
  
% Our channel matrix is nothing but a simple column vector.
H  = [h1; h2; h3];

% Solve for the MMSE Coefficients (The Wiener Hopf equation)
CT = Rx*H' * inv(H*Rx*H' + Rv) 

disp(['The estimator coefficients are: ' num2str(CT)]);

%% Transmitting a QPSK symbol stream
N    = 200;       % Number of TX Symbols
X    = sqrt(SigmaX) * (0.7071 * (2*randi([0,1], 1, N)-1) + 1j * 0.7071 * (2*randi([0,1], 1, N)-1)); 
 
% Generate complex Gaussian noise with the right noise variance SigmaV1, SigmaV2, and SigmaV3
V1   = sqrt(SigmaV1) * sqrt(1/2) * (randn(1, N) + 1j*randn(1, N));  
V2   = sqrt(SigmaV2) * sqrt(1/2) * (randn(1, N) + 1j*randn(1, N));
V3   = sqrt(SigmaV3) * sqrt(1/2) * (randn(1, N) + 1j*randn(1, N));

% The observation vectors
Y    = [X * h1 + V1; ...  % At receive antenna 1
        X * h2 + V2; ...  % At receive antenna 2
        X * h3 + V3];     % At receive antenna 3
    
% Compute the Estimate of the transmit symbol sequence x[n], as well as the associated error.
X_Estimate      = CT * Y; 
Error           = X_Estimate - X;               % The estimate of x[n] minus x[n]
SNR_X_Estimate  = var(X_Estimate)/var(X_Estimate - X);
disp(['The linear SNR of the estimate of X is equal to ' num2str(SNR_X_Estimate) ]);


figure(1);
subplot(2,2,1);
plot(real(Y(1,:)), imag(Y(1,:)), 'ko'); grid on; 
title('Observation Vector y1[n]');
xlabel('real'); ylabel('imag');  

subplot(2,2,2);
plot(real(Y(2,:)), imag(Y(2,:)), 'ko'); grid on; 
title('Observation Vector y2[n]');
xlabel('real'); ylabel('imag'); 

subplot(2,2,3);
plot(real(Y(3,:)), imag(Y(3,:)), 'ko'); grid on; 
title('Observation vector y3[n]');
xlabel('real'); ylabel('imag'); 

subplot(2,2,4);
plot(real(X_Estimate(1,:)), imag(X_Estimate(1,:)), 'ko'); grid on; 
title('The Final Estimate ');
xlabel('real'); ylabel('imag');axis([-1 1 -1 1])

