%% Text: Digital Signal Processing in Modern Communication Systems (Edition 2)
%% Chapter 8 - Section 4: Open Loop Spatial Muliplexing Code
%% Copyright: Andreas Schwarzinger Nov 2021
%%
%% Author: Andreas Schwarzinger                              Nov 2021

clc;
clear variables;
close all;

%% 0. We will use the MMSE estimator to detect the transmit symbol vector at the receiver.
%     Let's define some of the quantities we need for the estimator.
%     Define the signal and noise powers
SigmaX  =   1;  % The rms value of each transmit stream x1[n] and x2[n]
SigmaV1 = 0.1;  % The rms noise value at receive antenna 1
SigmaV2 = 0.1;  % The rms noise value at receive antenna 2
SigmaV3 = 0.1;  % The rms noise value at receive antenna 3
SigmaV4 = 0.1;  % The rms noise value at receive antenna 4

Rx = [SigmaX^2,        0; ...
             0, SigmaX^2];

%% 1. Let's define a channel with random coefficients
NumRepetitions              = 200;
N                           = 1000; % The number of input symbols for each run
Scenario                    = 1;    % Scenario 1,2,3 - indicate 2x2, 2x3 and 2x4 channel conditions
SNR_Array1                  = zeros(1, NumRepetitions);
SNR_Array2                  = zeros(2, NumRepetitions);

for Repetitions = 1:NumRepetitions
    % Generate QPSK streams for both x1[n] and x2[n]
    X             =      0.7071 * (2*randi([0,1], 2, N) - 1) ...
                    + 1j*0.7071 * (2*randi([0,1], 2, N) - 1);  
    
    % Generate complex Gaussian noise with noise variances SigmaV1, SigmaV2, SigmaV3 and Sigma4
    V1   = SigmaV1 * sqrt(1/2) * (randn(1, N) + 1j*randn(1, N));  
    V2   = SigmaV2 * sqrt(1/2) * (randn(1, N) + 1j*randn(1, N));
    V3   = SigmaV3 * sqrt(1/2) * (randn(1, N) + 1j*randn(1, N));
    V4   = SigmaV4 * sqrt(1/2) * (randn(1, N) + 1j*randn(1, N));       
                
    % Generate the channel matrix for the scenario
    switch(Scenario)
        case 1     % 2x2 channel matrix
            H  = (2*rand(2,2) - 1) + 1j*(2*rand(2,2) - 1); 
            Rv = [SigmaV1^2,         0; ...
                          0, SigmaV2^2];  
            V  = [V1; V2];
        case 2
            H = (2*rand(3,2) - 1) + 1j*(2*rand(3,2) - 1); 
            Rv = [SigmaV1^2,         0,          0; ...
                          0, SigmaV2^2,          0; ...
                          0,         0,  SigmaV3^2];
            V  = [V1; V2; V3];
        case 3
            H = (2*rand(4,2) - 1) + 1j*(2*rand(4,2) - 1);
            Rv = [SigmaV1^2,         0,        0,         0; ...
                          0, SigmaV2^2,        0,         0; ...
                          0,         0,  SigmaV3^2,       0; ...
                          0,         0,        0, SigmaV4^2];
            V  = [V1; V2; V3; V4];
        otherwise; error('Unsupported scenario.');
    end
    
    % Compute the Observation Vector Y
    Y = H*X + V;
    % Solve for the MMSE Coefficients (The Wiener Hopf equation)
    CT = Rx*H' * inv(H*Rx*H' + Rv); 

    % Compute the Estimate of the transmit vector X
    X_Estimate = CT * Y;
    
    % Compute the error and the SNR
    Error                      = X_Estimate - X;  
    SNR_X_Estimate1            = var(X_Estimate(1,:))/var(Error(1,:)); 
    SNR_X_Estimate2            = var(X_Estimate(2,:))/var(Error(2,:)); 
    SNR_Array2(:, Repetitions) = [SNR_X_Estimate1; SNR_X_Estimate2]; 
end

figure(1);
histogram(SNR_Array2(1,:), 10, 'facecolor', [1, 1, 1] );
xlabel('Linear SNR1'); axis([0 450 0 60]);


